from dexy.dexy_filter import DexyFilter
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.formatters.latex import LatexFormatter
from pygments.lexers.compiled import JavaLexer
import handlers.pexpect_handlers
import handlers.stdout_handlers
import json
import os
import platform
import subprocess

class JrubyHandler(handlers.stdout_handlers.ProcessStdoutHandler):
    VERSION = "jruby --version"
    EXECUTABLE = "jruby"
    INPUT_EXTENSIONS = [".rb"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['jruby']

class JirbHandler(handlers.pexpect_handlers.ProcessLinewiseInteractiveHandler):
    VERSION = "jirb --version"
    EXECUTABLE = "jirb --prompt-mode simple"
    INPUT_EXTENSIONS = [".rb"]
    PROMPT = ['>>', '?>']
    OUTPUT_EXTENSIONS = [".rbcon"]
    ALIASES = ['jirb']
    IGNORE_ERRORS = True

class JythonHandler(handlers.stdout_handlers.ProcessStdoutHandler):
    VERSION = "jython --version"
    EXECUTABLE = "jython"
    INPUT_EXTENSIONS = [".py"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['jython']

    @classmethod
    def enabled(self):
        if platform.system() in ('Linux', 'Windows'):
            return True
        elif platform.system() in ('Darwin'):
            print "The jython dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled."
            return False
        else:
            print """Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jython dexy filter should not be run on MacOS due to a serious bug.""" % platform.system()
            return True

class JythonInteractiveHandler(handlers.pexpect_handlers.ProcessLinewiseInteractiveHandler):
    VERSION = "jython --version"
    EXECUTABLE = "jython -i"
    INPUT_EXTENSIONS = [".py", ".txt"]
    OUTPUT_EXTENSIONS = [".pycon"]
    ALIASES = ['jythoni']

    @classmethod
    def enabled(self):
        if platform.system() in ('Linux', 'Windows'):
            return True
        elif platform.system() in ('Darwin'):
            print "The jythoni dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled."
            return False
        else:
            print """Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jythoni dexy filter should not be run on MacOS due to a serious bug.""" % platform.system()
            return True

class JavaHandler(DexyFilter):
    EXECUTABLE = "javac"
    VERSION = "java -version"
    INPUT_EXTENSIONS = [".java"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['java']

    def process(self):
        if self.artifact.args.has_key('main'):
            main_method = self.artifact.args['main']
        else:
            main_method = os.path.splitext(os.path.basename(self.artifact.name))[0]

        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None

        # Create a temp dir to hold source + compiled files, since we need .java files
        # named with their original names.
        self.artifact.create_temp_dir()
        wf = os.path.join(self.artifact.temp_dir(), os.path.basename(self.artifact.name))
        f = open(wf, "w")
        f.write(self.artifact.input_text())
        f.close()

        if env.has_key("CLASSPATH"):
            # need to add 1 level to classpath since we are working in a subdir of artifacts/
            env['CLASSPATH'] = ":".join(["../%s" % x for x in env['CLASSPATH'].split(":")])
        else:
            env['CLASSPATH'] = None

        cp = "."
        if env['CLASSPATH']:
                cp = "%s:%s" % (cp, env['CLASSPATH'])

        cwd = self.artifact.temp_dir()

        command = "javac -classpath %s %s" % (cp, os.path.basename(self.artifact.name))
        self.log.debug(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=cwd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=env)
        stdout, stderr = proc.communicate()
        self.log.debug("stdout from compiler:\n%s" % stdout)
        self.log.debug("stderr from compiler:\n%s" % stderr)
        if proc.returncode > 0:
            raise Exception("a problem occurred running %s.\ndetails:\n%s" % (
                command, stderr))


        command = "java -cp %s %s" % (cp, main_method)
        self.log.debug(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=cwd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=env)
        stdout, stderr = proc.communicate()
        if proc.returncode > 0:
            raise Exception("a problem occurred running %s.\ndetails:\n%s" % (
                command, stderr))

        self.artifact.data_dict['1'] = stdout

class JavadocsJsonFilter(DexyFilter):
    ALIASES = ['javadoc', 'javadocs']

    def nested_subclasses(self, j, qualified_class_name, nest=None, indent = 0):
        if not nest:
            nest = [[indent, qualified_class_name]]
        package_name, x, class_name = qualified_class_name.rpartition(".")
        if j['packages'][package_name]['classes'][class_name].has_key('subclasses'):
            subclasses = j['packages'][package_name]['classes'][class_name]['subclasses']
            for qualified_subclass_name in subclasses:
                nest.append([indent+1, qualified_subclass_name])
                subclass_package_name, x, subclass_name = qualified_subclass_name.rpartition(".")
                if j['packages'][subclass_package_name]['classes'][subclass_name].has_key('subclasses'):
                    self.nested_subclasses(j, qualified_subclass_name, nest, indent+1)
        return nest

    def process_text(self, input_text):
        j = json.loads(input_text)

        html_formatter = HtmlFormatter()
        latex_formatter = LatexFormatter()
        lexer = JavaLexer()

        # Update our JSON array with each input data file.
        def update_dict(o, n):
            for k in o.keys():
                if n.has_key(k):
                    v = o[k]
                    if isinstance(v, dict):
                        o[k] = update_dict(v, n[k])
                    else:
                        o[k].update(n[k])
            return o

        for k, a in self.artifact.inputs().items():
            new_data = json.loads(a.output_text())
            j = update_dict(j, new_data)

        for p in j['packages']:
            for k in j['packages'][p]['classes'].keys():

                klass = j['packages'][p]['classes'][k]
                if klass.has_key('superclass'):
                    superclass_package, _, superclass_name = klass['superclass'].rpartition(".")
                    if j['packages'].has_key(superclass_package):
                        if not j['packages'][superclass_package]['classes'].has_key(superclass_name):
                            print "Can't find", superclass_name, "in package", superclass_package
                        else:
                            superclass = j['packages'][superclass_package]['classes'][superclass_name]
                            if not superclass.has_key('subclasses'):
                                superclass['subclasses'] = []

                            superclass['subclasses'].append("%s.%s" % (p, k))

                if klass.has_key('interfaces'):
                    for iface in klass['interfaces']:
                        iface_package, _, iface_name = iface.rpartition(".")
                        if j['packages'].has_key(iface_package):
                            if not j['packages'][iface_package]['classes'].has_key(iface_name):
                                print "Can't find", iface_name, "in package", iface_package
                            else:
                                iface_dict = j['packages'][iface_package]['classes'][iface_name]
                                if not iface_dict.has_key('implementers'):
                                    iface_dict['implementers'] = []
                                iface_dict['implementers'].append(klass['qualified-name'])
                if klass.has_key('source') and klass['source']:
                    source = klass['source']
                    j['packages'][p]['classes'][k]['source-html'] = str(highlight(source, lexer, html_formatter))
                    j['packages'][p]['classes'][k]['source-latex'] = str(highlight(source, lexer, latex_formatter))

                for m in j['packages'][p]['classes'][k]['methods'].keys():
                    source = j['packages'][p]['classes'][k]['methods'][m]['source']
                    if not source:
                        source = ""

                    # TODO - try running comment text through Textile
                    # interpreter for HTML and LaTeX options
                    comment = j['packages'][p]['classes'][k]['methods'][m]['comment-text']


                    j['packages'][p]['classes'][k]['methods'][m]['comment'] = comment
                    j['packages'][p]['classes'][k]['methods'][m]['source'] = source
                    j['packages'][p]['classes'][k]['methods'][m]['source-html'] = str(highlight(source, lexer, html_formatter))
                    j['packages'][p]['classes'][k]['methods'][m]['source-latex'] = str(highlight(source, lexer, latex_formatter))

                for c in j['packages'][p]['classes'][k]['constructors'].keys():
                    source = j['packages'][p]['classes'][k]['constructors'][c]['source']
                    if not source:
                        source = ""

                    # TODO - try running comment text through Textile
                    # interpreter for HTML and LaTeX options
                    comment = j['packages'][p]['classes'][k]['constructors'][c]['comment-text']

                    j['packages'][p]['classes'][k]['constructors'][c]['comment'] = comment
                    j['packages'][p]['classes'][k]['constructors'][c]['source'] = source
                    j['packages'][p]['classes'][k]['constructors'][c]['source-html'] = str(highlight(source, lexer, html_formatter))
                    j['packages'][p]['classes'][k]['constructors'][c]['source-latex'] = str(highlight(source, lexer, latex_formatter))

        return json.dumps(j)



# Original topological sort code written by Ofer Faigon
# (www.bitformation.com) and used with permission

def topological_sort(items, partial_order):
    """
    Perform topological sort. items is a list of items to be sorted.
    partial_order is a list of pairs. If pair (a,b) is in it, it means that item
    a should appear before item b. Returns a list of the items in one of the
    possible orders, or None if partial_order contains a loop.
    """

    def add_node(graph, node):
        """Add a node to the graph if not already exists."""
        if not graph.has_key(node):
            graph[node] = [0] # 0 = number of arcs coming into this node. 

    def add_arc(graph, fromnode, tonode):
        """
        Add an arc to a graph. Can create multiple arcs.
        The end nodes must already exist.
        """
        graph[fromnode].append(tonode)
        graph[tonode][0] = graph[tonode][0] + 1

    # step 1 - create a directed graph with an arc a->b for each input 
    # pair (a,b). 
    # The graph is represented by a dictionary. The dictionary contains 
    # a pair item:list for each node in the graph. /item/ is the value 
    # of the node. /list/'s 1st item is the count of incoming arcs, and 
    # the rest are the destinations of the outgoing arcs. For example: 
    # {'a':[0,'b','c'], 'b':[1], 'c':[1]} 
    # represents the graph: c <-- a --> b 
    # The graph may contain loops and multiple arcs. 
    # Note that our representation does not contain reference loops to 
    # cause GC problems even when the represented graph contains loops, 
    # because we keep the node names rather than references to the nodes. 
    graph = {}
    for v in items:
        add_node(graph, v)
    for a, b in partial_order:
        add_arc(graph, a, b)
    
    # Step 2 - find all roots (nodes with zero incoming arcs).
    roots = [node for (node, nodeinfo) in graph.items() if nodeinfo[0]==0]

    # step 3 - repeatedly emit a root and remove it from the graph. Removing 
    # a node may convert some of the node's direct children into roots. 
    # Whenever that happens, we append the new roots to the list of 
    # current roots.
    sorted = []
    while len(roots) != 0:
        # If len(roots) is always 1 when we get here, it means that 
        # the input describes a complete ordering and there is only 
        # one possible output. 
        # When len(roots) > 1, we can choose any root to send to the 
        # output; this freedom represents the multiple complete orderings 
        # that satisfy the input restrictions. We arbitrarily take one of 
        # the roots using pop(). Note that for the algorithm to be efficient, 
        # this operation must be done in O(1) time.
        root = roots.pop()
        sorted.append(root)
        for child in graph[root][1:]:
            graph[child][0] = graph[child][0] - 1
            if graph[child][0] == 0:
                roots.append(child)
        del graph[root]

    if len(graph.items()) != 0:
        # There is a loop in the input
        return None
    return sorted



from dexy.dexy_filter import DexyFilter
from ordereddict import OrderedDict
import simplejson as json
import os
import pexpect
import re
import sys
import time

class ProcessLinewiseInteractiveHandler(DexyFilter):
    """
    Intended for use with interactive processes, such as python interpreter,
    where your goal is to have a session transcript divided into same sections
    as input. Sends input line-by-line.
    """
    PROMPT = ['>>>', '...'] # Python uses >>> prompt normally and ... when in multi-line structures like loops
    TRIM_PROMPT = '>>>'
    LINE_ENDING = "\r\n"
    IGNORE_ERRORS = False # Allow overriding default per-handler.
    SAVE_VARS_TO_JSON_CMD = None
    ALIASES = None

    def process_dict(self, input_dict):
        output_dict = OrderedDict()

        if self.artifact.args.has_key('timeout'):
            timeout = self.artifact.args['timeout']
            self.log.info("using custom timeout %s for %s" % (timeout, self.artifact.key))
        else:
            timeout = None
        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
            self.log.info("adding to env: %s" % self.artifact.args['env'])
        else:
            env = None

        proc = pexpect.spawn(self.EXECUTABLE, cwd=self.artifact.artifacts_dir, env=env)
        proc.expect_exact(self.PROMPT, timeout=timeout)
        start = (proc.before + proc.after)

        search_terms = ["%s%s" % (self.LINE_ENDING, p) for p in self.PROMPT]

        for k, s in input_dict.items():
            # TODO Should stop processing if an error is raised.
            section_transcript = start
            start = ""
            for l in (s+"\r\n").splitlines():
                section_transcript += start
                start = ""
                proc.sendline(l)
                proc.expect_exact(search_terms, timeout=timeout)
                section_transcript += proc.before
                start = proc.after

            # Strip blank lines/trailing prompts at end of section
            lines = section_transcript.split(self.LINE_ENDING)
            while len(lines) > 0 and re.match("^\s*(%s)\s*$|^\s*$" % self.TRIM_PROMPT, lines[-1]):
                lines = lines[0:-1]

#            output_dict[k] = ''.join([c for c in self.LINE_ENDING.join(lines) if ord(c) > 31 or ord(c) in [9, 10]])
            output_dict[k] = self.LINE_ENDING.join(lines)

        record_vars = self.artifact.args.has_key('record_vars') and self.artifact.args['record_vars']
        if record_vars:
            if not self.SAVE_VARS_TO_JSON_CMD:
                raise Exception("Can't record vars since SAVE_VARS_TO_JSON_CMD not set.")
            artifact = self.artifact.add_additional_artifact(self.artifact.key + "-vars", 'json')
            cmd = self.SAVE_VARS_TO_JSON_CMD % artifact.filename()

            section_transcript = start
            start = ""
            for l in cmd.splitlines():
                section_transcript += start
                start = ""
                proc.sendline(l)
                proc.expect_exact(search_terms, timeout=timeout)
                section_transcript += proc.before
                start = proc.after

            output_dict['dexy--save-vars'] = section_transcript

        try:
            proc.close()
        except pexpect.ExceptionPexpect:
            print "process %s may not have closed" % proc.pid

        if proc.exitstatus is not None and proc.exitstatus != 0:
            if not (self.IGNORE_ERRORS or self.artifact.args.ignore_errors):
                raise Exception("""proc returned nonzero status code! if you don't
want dexy to raise errors on failed scripts then pass the --ignore-errors option""")
        return output_dict

class PythonLinewiseInteractiveHandler(ProcessLinewiseInteractiveHandler):
    EXECUTABLE = 'python'
    VERSION = 'python --version'
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".pycon"]
    ALIASES = ['pycon']
    SAVE_VARS_TO_JSON_CMD = """dexy__vars_file = open("%s", "w")
dexy__x = {}
for dexy__k, dexy__v in locals().items():
    dexy__x[dexy__k] = str(dexy__v)

json.dump(dexy__x, dexy__vars_file)
dexy__vars_file.close()
"""

class RLinewiseInteractiveHandler(ProcessLinewiseInteractiveHandler):
    """
    Runs R
    """
    EXECUTABLE = "R --quiet --vanilla"
    VERSION = "R --version"
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = ['.Rout']
    PROMPT = [">", "+"]
    TRIM_PROMPT = ">"
    ALIASES = ['r', 'rint']
    SAVE_VARS_TO_JSON_CMD = """
if ("rjson" %%in%% installed.packages()) {
    library(rjson)
    dexy__json_file <- file("%s", "w")
    writeLines(toJSON(as.list(environment())), dexy__json_file)
    close(dexy__json_file)
} else {
   cat("Can't automatically save environment to JSON since rjson package not installed.")
}
"""

class RhinoInteractiveHandler(ProcessLinewiseInteractiveHandler):
    """
    Runs rhino JavaScript interpeter.
    """
    EXECUTABLE = "rhino"
    INPUT_EXTENSIONS = [".js"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['jsint', 'rhino']
    PROMPT = "js> "

class ClojureInteractiveHandler(ProcessLinewiseInteractiveHandler):
    """
    Runs clojure.
    """
    EXECUTABLE = None
    EXECUTABLES = ['clojure', 'clj -r', 'java clojure.main']
    INPUT_EXTENSIONS = [".clj"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['clj', 'cljint']
    PROMPT = "user=> "

class ProcessTimingHandler(DexyFilter):
    """
    Runs python code N times and reports timings.
    """
    EXECUTABLE = 'python'
    VERSION = 'python --version'
    N = 10
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".times"]
    ALIASES = ['timing', 'pytime']

    def process(self):
        self.artifact.generate_workfile()
        times = []
        for i in xrange(self.N):
            start = time.time()
            pexpect.run("%s %s" % (self.EXECUTABLE, self.artifact.work_filename()))
            times.append("%s" % (time.time() - start))
        self.artifact.data_dict['1'] = "\n".join(times)

class Pdf2Jpg(DexyFilter):
    """
    Converts a PDF file to JPEG format.
    """
    INPUT_EXTENSIONS = ['.pdf']
    OUTPUT_EXTENSIONS = ['.jpg']
    ALIASES = ['pdf2jpg', 'pdf2jpeg']
    EXECUTABLE = "gs"
    VERSION = "gs --version"
    BINARY = True

    def process(self):
        # Can't use generate_workfile as input is binary, not text-based.
        # TODO should we be doing this in general rather than generating workfiles?
        wf = self.artifact.previous_artifact_filename
        of = self.artifact.filename()
        command = "%s -dSAFER -dNOPAUSE -dBATCH -sDEVICE=jpeg -sOutputFile=%s ../%s" % (self.executable(), of, wf)
        self.log.debug(command)
        self.artifact.stdout = pexpect.run(command, cwd=self.artifact.artifacts_dir)

class Pdf2Png(DexyFilter):
    """
    Converts a PDF file to PNG format.
    """
    INPUT_EXTENSIONS = ['.pdf']
    OUTPUT_EXTENSIONS = ['.png']
    ALIASES = ['pdf2png']
    EXECUTABLE = "gs"
    VERSION = "gs --version"
    BINARY = True

    def process(self):
        # Can't use generate_workfile as input is binary, not text-based.
        # TODO should we be doing this in general rather than generating workfiles?
        wf = self.artifact.previous_artifact_filename
        of = self.artifact.filename()
        command = "%s -dSAFER -dNOPAUSE -dBATCH -sDEVICE=png16m -r300 -sOutputFile=%s %s" % (self.executable(), of, wf)
        self.log.debug(command)
        self.artifact.stdout = pexpect.run(command, cwd=self.artifact.artifacts_dir)

class Ps2Pdf(DexyFilter):
    """
    Converts a PS file to PDF format.
    """
    INPUT_EXTENSIONS = [".ps", ".txt"]
    OUTPUT_EXTENSIONS = [".pdf"]
    ALIASES = ['ps2pdf']
    EXECUTABLE = 'ps2pdf'
    BINARY = True

    def process(self):
        self.artifact.generate_workfile()
        wf = self.artifact.work_filename()
        of = self.artifact.filename()
        command = "%s %s %s" % (self.executable(), wf, of)
        self.log.debug(command)
        self.artifact.stdout = pexpect.run(command, cwd=self.artifact.artifacts_dir)




from dexy.dexy_filter import DexyFilter
import os
import re
import subprocess

class RagelRubyHandler(DexyFilter):
    """
    Runs ragel for ruby.
    """
    INPUT_EXTENSIONS = [".rl"]
    OUTPUT_EXTENSIONS = [".rb"]
    ALIASES = ['rlrb', 'ragelruby']
    VERSION = 'ragel --version'
    EXECUTABLE = 'ragel -R'

    def process(self):
        self.artifact.generate_workfile()
        wf = self.artifact.work_filename()
        command = "%s %s -o %s" % (self.executable(), wf, self.artifact.filename())

        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None

        self.log.info("running: %s" % command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)

        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

class BibHandler(DexyFilter):
    INPUT_EXTENSIONS = [".tex"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['bib']

    """Filter which replaces a hard-coded bibtex file with a .bib file specified as an input."""
    def process_text(self, input_text):
        for k, a in self.artifact.inputs().items():
            if a.filename().endswith("bib"):
                bib_file_basename = os.path.splitext(os.path.basename(a.name))[0]

                input_text = re.sub("bibliography{[^}]+}", "bibliography{%s}" % a.filename(), input_text)

        return input_text

class LatexHandler(DexyFilter):
    """
    Generates a PDF file from LaTeX source.
    """
    INPUT_EXTENSIONS = [".tex", ".txt"]
    OUTPUT_EXTENSIONS = [".pdf", ".png"]
    EXECUTABLES = ['pdflatex', 'latex']
    ALIASES = ['latex']
    BINARY = True
    FINAL = True

    def process(self):
        latex_filename = self.artifact.filename().replace(self.artifact.ext, ".tex")
        self.artifact.generate_workfile(latex_filename)

        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None

        latex_command = "%s -interaction=batchmode %s" % (self.__class__.executable(), latex_filename)

        if self.__class__.executable_present("bibtex"):
            bibtex_command = "bibtex %s" % os.path.splitext(self.artifact.filename())[0]
        else:
            bibtex_command = None

        self.artifact.stdout = ""
        def run_cmd(command):
            self.log.info("running: %s" % command)
            proc = subprocess.Popen(command, shell=True,
                                    cwd=self.artifact.artifacts_dir,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)

            stdout, stderr = proc.communicate()
            self.artifact.stdout += stdout
            if proc.returncode > 2: # Set at 2 for now as this is highest I've hit, better to detect whether PDF has been generated?
                raise Exception("latex error, look for information in %s" %
                                latex_filename.replace(".tex", ".log"))
            elif proc.returncode > 0:
                print """A non-critical latex error has occurred running %s,
                status code returned was %s, look for information in %s""" % (
                self.artifact.key, proc.returncode,
                latex_filename.replace(".tex", ".log"))


        runbibtex = bibtex_command # TODO allow opting out of running bibtex in args
        if runbibtex:
            run_cmd(latex_command) #generate aux
            run_cmd(bibtex_command) #generate bbl
        run_cmd(latex_command) #first run
        run_cmd(latex_command) #second run - fix references


class EmbedFonts(DexyFilter):
    INPUT_EXTENSIONS = [".pdf"]
    OUTPUT_EXTENSIONS = [".pdf"]
    EXECUTABLE = 'ps2pdf'
    ALIASES = ['embedfonts', 'prepress']
    FINAL = True
    BINARY = True

    def process(self):
        pf = self.artifact.previous_artifact_filename
        af = self.artifact.filename()

        env = None

        command = "%s -dPDFSETTINGS=/prepress %s %s" % (self.EXECUTABLE, pf, af)
        self.log.info(command)

        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)

        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

        command = "%s %s" % ("pdffonts", af)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)

        stdout, stderr = proc.communicate()
        self.artifact.stdout += stdout

class Rd2PdfHandler(DexyFilter):
    INPUT_EXTENSIONS = [".Rd"]
    OUTPUT_EXTENSIONS = [".pdf", ".dvi"]
    EXECUTABLE = 'R CMD Rd2pdf'
    VERSION = 'R CMD Rd2pdf -v'
    ALIASES = ['rd2pdf', 'Rd2pdf']
    FINAL = True

    def process(self):
        self.artifact.generate_workfile()
        wf = self.artifact.work_filename()
        af = self.artifact.filename()
        title = os.path.splitext(self.artifact.name)[0].replace("_", " ")
        command = "%s --output=%s --title=\"%s\" %s" % (self.executable(), af,
                                                    title, wf)
        self.log.info(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

class RBatchHandler(DexyFilter):
    """Runs R code in batch mode."""
    EXECUTABLE = 'R CMD BATCH --vanilla --quiet --no-timing'
    VERSION = "R --version"
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['rintbatch']

    def process(self):
        self.artifact.generate_workfile()
        work_file = os.path.basename(self.artifact.work_filename())
        artifact_file = os.path.basename(self.artifact.filename())
        command = "%s %s %s" % (self.EXECUTABLE, work_file, artifact_file)
        self.log.info(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout
        self.artifact.set_data_from_artifact()

class ROutputBatchHandler(DexyFilter):
    """Runs R code in batch mode. Uses the --slave flag so doesn't echo commands, just returns output."""
    EXECUTABLE = 'R CMD BATCH --vanilla --quiet --slave --no-timing'
    VERSION = "R --version"
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['routbatch']

    def process(self):
        self.artifact.generate_workfile()
        work_file = os.path.basename(self.artifact.work_filename())
        artifact_file = os.path.basename(self.artifact.filename())
        command = "%s %s %s" % (self.EXECUTABLE, work_file, artifact_file)
        self.log.info(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout
        self.artifact.set_data_from_artifact()

class DotHandler(DexyFilter):
    """
    Renders .dot files to either PNG or PDF images.
    """
    INPUT_EXTENSIONS = [".dot"]
    OUTPUT_EXTENSIONS = [".png", ".pdf"]
    EXECUTABLE = 'dot'
    VERSION = 'dot -V'
    ALIASES = ['dot', 'graphviz']
    FINAL = True
    BINARY = True

    def process(self):
        self.artifact.generate_workfile()
        wf = self.artifact.work_filename()
        af = self.artifact.filename()
        ex = self.artifact.ext.replace(".", "")
        command = "%s -T%s -o%s %s" % (self.executable(), ex, af, wf)
        self.log.info(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

class RubyInteractiveHandler(DexyFilter):
    """Run Ruby, takign input from input files."""
    VERSION = "ruby --version"
    EXECUTABLE = "ruby"
    INPUT_EXTENSIONS = [".rb"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['rbint']

    def process(self):
        self.artifact.generate_workfile()
        wf = self.artifact.work_filename()
        command = "%s %s" % (self.EXECUTABLE, wf)
        self.log.debug(command)
        for k, a in self.artifact.inputs().items():
            for s, t in a.data_dict.items():
                proc = subprocess.Popen(command, shell=True,
                                        cwd=self.artifact.artifacts_dir,
                                        stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                       )
                stdout, stderr = proc.communicate(t)
                self.artifact.data_dict[s] = stdout
                self.artifact.stdout = stdout + "\n" + stderr




try:
    import web
    import sqlite3
#    DB_AVAILABLE = True
    DB_AVAILABLE = False # Disable this for now.
    web.config.debug = False
except ImportError:
    DB_AVAILABLE = False

class Db(object):
    DATABASE_FILE = 'testdb'
    DATABASE_TYPE = 'sqlite'

    def __init__(self):
        self.conn = web.database(dbn=self.DATABASE_TYPE, db=self.DATABASE_FILE)

        try:
            self.conn.query("""CREATE TABLE tasks (
                artifact_hashstring text,
                document_key text,
                batch_id integer,
                batch_order integer,
                host text,
                pid integer,
                exit_status integer,
                complete boolean default 'f'
            );""")
        except sqlite3.OperationalError:
            # In this case, the db already exists.
            pass

    def next_batch_id(self):
        # TODO Replace with UUID?
        old_batch_id = self.conn.select("tasks", what="max(batch_id)")[0]["max(batch_id)"]
        if not old_batch_id:
            old_batch_id = 0
        return old_batch_id + 1

    def next_batch_order(self, batch_id):
        old_batch_order = self.conn.select("tasks", what="max(batch_order)", where="batch_id=%s" % batch_id)[0]["max(batch_order)"]
        if not old_batch_order:
            old_batch_order = 0 # Need to start at 0, not -1, since 0 evals to false.
        return old_batch_order + 1

    def insert_artifact(self, artifact, batch_id):
        self.conn.insert("tasks",
            batch_id=batch_id,
            batch_order=self.next_batch_order(batch_id),
            document_key=artifact.name,
            artifact_hashstring = artifact.hashstring
        )



from dexy.dexy_filter import DexyFilter
import subprocess
import time

### @export "clang-handler"
class ClangHandler(DexyFilter):
    """Compiles C code using clang compiler, then runs compiled program."""
    VERSION = "clang --version"
    EXECUTABLE = "clang"
    INPUT_EXTENSIONS = [".c"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['clang']

    def process(self):
        self.artifact.generate_workfile()
        of = self.artifact.temp_filename(".o")
        wf = self.artifact.work_filename()
        command = "%s %s -o %s" % (self.EXECUTABLE, wf, of)
        self.log.debug(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                )
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

        command = "./%s" % of
        self.log.debug(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                )
        stdout, stderr = proc.communicate()
        self.artifact.data_dict['1'] = stdout
        self.artifact.stdout += stderr

### @export "clang-interactive-handler"
class ClangInteractiveHandler(DexyFilter):
    """Compiles C code using clang compiler, then runs compiled program, reading
    input from any input files."""
    VERSION = "clang --version"
    EXECUTABLE = "clang"
    INPUT_EXTENSIONS = [".c"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['cint']

    def process(self):
        self.artifact.generate_workfile()
        of = self.artifact.temp_filename(".o")
        wf = self.artifact.work_filename()
        command = "%s %s -o %s" % (self.EXECUTABLE, wf, of)
        self.log.debug(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                )
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

        command = "./%s" % of
        self.log.debug(command)
        for k, a in self.artifact.inputs().items():
            for s, t in a.data_dict.items():
                proc = subprocess.Popen(command, shell=True,
                                        cwd=self.artifact.artifacts_dir,
                                        stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                       )
                stdout, stderr = proc.communicate(t)
                self.artifact.data_dict[s] = stdout
                self.artifact.stdout = self.artifact.stdout + "\n" + stderr

### @export "clang-timing-handler"
class ClangTimingHandler(DexyFilter):
    """Compiles C code using clang compiler, then runs compiled program N times
    reporting timings."""
    N = 10
    INPUT_EXTENSIONS = [".txt", ".c"]
    OUTPUT_EXTENSIONS = [".times"]
    ALIASES = ['ctime']
    EXECUTABLE = 'clang'
    VERSION = "clang --version"

    def process(self):
        self.artifact.generate_workfile()
        of = self.artifact.temp_filename(".o")
        wf = self.artifact.work_filename()
        command = "%s %s -o %s" % (self.EXECUTABLE, wf, of)
        self.log.debug(command)
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                )
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

        command = "./%s" % of
        self.log.debug(command)
        times = []
        for i in xrange(self.N):
            start = time.time()
            proc = subprocess.Popen(command, shell=True,
                                    cwd=self.artifact.artifacts_dir,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    )
            stdout, stderr = proc.communicate()
            times.append("%s" % (time.time() - start))
        self.artifact.data_dict['1'] = "\n".join(times)

### @export "cpp-handler"
class CppHandler(ClangHandler):
    """Compiles and then runs C++ code."""
    VERSION = "c++ --version"
    EXECUTABLE ="c++"
    INPUT_EXTENSIONS = [".cpp"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['cpp']

### @export "gcc-handler"
class CHandler(ClangHandler):
    """Compiles C code using gcc compiler, then runs compiled program."""
    VERSION = "gcc --version"
    EXECUTABLE = "gcc"
    INPUT_EXTENSIONS = [".c"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['c', 'gcc']





from dexy.dexy_filter import DexyFilter
from dexy.document import Document
from dexy.reporter import Reporter
from dexy.topological_sort import topological_sort
from dexy.utils import AttrDict
from inspect import isclass
from ordereddict import OrderedDict
import dexy.db
import fnmatch
import glob
import json
import logging
import os
import pprint
import re
import sqlite3
import sre_constants
import sys

class Controller(object):
    def __init__(self):
        self.config = {}

        # In actual usage, these initialized values are replaced by optparse
        # Namespace or the argparse equivalent. These initialized values are
        # used in testing, AttrDict is used so we can simulate the globals
        # attr.
        self.args = AttrDict({})
        self.args.globals = {}

        self.install_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
        if not hasattr(self, 'log'):
            self.log = logging.getLogger('dexy')
            self.log.propagate = 0 # Stops logs being written to STDOUT if another library redirects root logger.

    def find_reporters(self):
        d1 = os.path.abspath(os.path.join(self.install_dir, 'reporters'))
        d2 = os.path.abspath(os.path.join(os.curdir, 'reporters'))

        if d1 == d2:
            reporter_dirs = [d1]
        else:
            reporter_dirs = [d1,d2]

        reporters = []

        for d in reporter_dirs:
            if os.path.exists(d):
                for f in os.listdir(d):
                    if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
                        self.log.info("Loading reporters in %s" % os.path.join(d, f))
                        basename = f.replace(".py", "")
                        module = "reporters.%s" % basename

                        try:
                            __import__(module)
                        except ImportError as e:
                            self.log.warn("reporters defined in %s are not available: %s" % (module, e))

                        if not sys.modules.has_key(module):
                            continue

                        mod = sys.modules[module]

                        for k in dir(mod):
                            klass = mod.__dict__[k]
                            if isclass(klass) and not (klass == Reporter) and issubclass(klass, Reporter):
                                reporters.append(klass)
        self.reports_dirs = [r.REPORTS_DIR for r in reporters]
        return reporters

    def find_filters(self):
        sys.path.append(os.curdir)

        handler_dirs = []
        # Need to give local directory (in which we place custom per-project filters)
        # a different name or else Python can't see modules in that dir.
        # This handler/filter switching is confusing, should be improved.
        h1 = os.path.abspath(os.path.join(self.install_dir, 'handlers'))
        h2 = os.path.abspath(os.path.join(os.curdir, 'filters'))

        for h in [h1, h2]:
            if os.path.exists(h) and not h in handler_dirs:
                handler_dirs.append(h)

        handlers = {}

        for a in DexyFilter.ALIASES:
            handlers[a] = DexyFilter

        for d in handler_dirs:
            self.log.info("Automatically loading all filters found in %s" % d)
            for f in os.listdir(d):
                if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
                    self.log.info("Loading filters in %s" % os.path.join(d, f))
                    basename = f.replace(".py", "")
                    if d.endswith('handlers'):
                        module = "handlers.%s" % basename
                    elif d.endswith('filters'):
                        module = "filters.%s" % basename
                    else:
                        raise Exception(d)

                    try:
                        __import__(module)
                    except ImportError as e:
                        self.log.warn("filters defined in %s are not available: %s" % (module, e))

                    if not sys.modules.has_key(module):
                        continue

                    mod = sys.modules[module]

                    for k in dir(mod):
                        klass = mod.__dict__[k]
                        if isclass(klass) and not (klass == DexyFilter) and issubclass(klass, DexyFilter):
                            if not klass.ALIASES:
                                self.log.info("class %s is not available because it has no aliases" % klass.__name__)
                            elif not klass.executable_present():
                                self.log.info("class %s is not available because %s not found" %
                                              (klass.__name__, klass.executable()))
                            elif not klass.enabled():
                                self.log.info("class %s is not available because it is not enabled" %
                                              (klass.__name__))
                            else:
                                for a in klass.ALIASES:
                                    if handlers.has_key(a):
                                        raise Exception("duplicate key %s called from %s in %s" % (a, k, f))
                                    handlers[a] = klass
                                    self.log.info("registered alias %s for class %s" % (a, k))
            self.log.info("...finished loading filters from %s" % d)
        return handlers

    def register_filters(self):
        self._filters = self.find_filters()

    def get_handler_for_alias(self, alias):
        """Given an alias, return the corresponding handler."""
        if self._filters.has_key(alias):
            return self._filters[alias]
        elif alias.startswith("-") or alias.startswith("alias-") or alias.startswith("al-") or alias in ['al', 'alias']:
            self._filters[alias] = DexyFilter
        else:
            raise Exception("You requested filter alias '%s' but this is not available." % alias)

    def register_reporters(self):
        self.reporters = self.find_reporters()

    def load_config(self, path_to_dir, rel_to=os.curdir):
        # This looks in every parent directory for a config file and combines them.
        relative_path = os.path.relpath(path_to_dir, rel_to)
        path_elements = relative_path.split(os.sep)
        config_dict = {}
        global_args = {}
        for i in range(0,len(path_elements)+1):
            config_file_path_elements = [rel_to] + path_elements[0:i] + [self.config_file]
            config_filename = os.path.join(*config_file_path_elements)
            config_files = glob.glob(config_filename)
            for cf in config_files:
                self.log.info("loading config %s" % cf)
                config_file = open(cf, "r")
                try:
                    json_dict = json.load(config_file)
                except ValueError as e:
                    raise Exception("""Your config file %s has invalid JSON. Details: %s""" %
                                    (cf, e.message))
                if json_dict.has_key("$reset"):
                    config_dict = json_dict
                else:
                    config_dict.update(json_dict)

                if json_dict.has_key("$globals"):
                    global_args.update(json_dict["$globals"])
                config_dict['$globals'] = global_args
        self.config[path_to_dir] = config_dict

    def process_config(self):
        def parse_doc(path, input_directive, args = {}):
            # If a specification is nested in a dependency, then input_directive
            # may be a dict. If so, split it into parts before continuing.
            try:
                a, b = input_directive.popitem()
                input_directive = a
                args = b
            except AttributeError:
                pass

            tokens = input_directive.split("|")
            if "/" in tokens[0]:
                raise Exception("paths not allowed in tokens: %s" % tokens[0])
            if path == '.':
                glob_string = tokens[0]
            else:
                glob_string = os.path.join(re.sub("^\./", "", path), tokens[0])
            filters = tokens[1:]

            docs = []

            # virtual document
            if re.search("@", glob_string):
                virtual = True
                # TODO some virtual files are local, not remote. test on
                # presence of 'url' or something more appropriate.
                if not self.allow_remote:
                    raise Exception("""You are attempting to access a remote file.
                                    You must enable --dangerous mode to do this.
                                    Please check Dexy help and call the dexy
                                    command again.""")
                glob_string = glob_string.replace("@", "")
            else:
                virtual = False

            regex = fnmatch.translate(glob_string).replace(".*", "(.*)")
            matcher = re.compile(regex)

            files = glob.glob(glob_string)

            nofiles = len(files) == 0

            if nofiles and virtual:
                files = [glob_string]

            for f in files:
                create = True
                if not virtual:
                    if os.path.isdir(f):
                        create = False

                if args.has_key('disabled'):
                    if args['disabled']:
                        create = False
                        print "document %s|%s disabled" % (f, "|".join(filters))

                inputs = []
                if args.has_key('inputs'):
                    if isinstance(args['inputs'], str):
                        raise Exception("""this input should be an array,
                                        not a string: %s""" % args['inputs'])
                    for i in args['inputs']:
                        for doc in parse_doc(path, i):
                            inputs.append(doc.key())
                m = matcher.match(f)
                if m and len(m.groups()) > 0:
                    rootname = matcher.match(f).group(1)

                # The 'ifinput' directive says that if an input exists matching
                # the specified pattern, we should create this document and it
                # will depend on the specified input.
                if args.has_key('ifinput'):
                    self.log.debug(f)
                    if isinstance(args['ifinput'], str) or isinstance(args['ifinput'], unicode):
                        ifinputs = [args['ifinput']]
                    else:
                        self.log.debug("treating input %s as iterable. class: %s" % (
                            args['ifinput'], args['ifinput'].__class__.__name__))
                        ifinputs = args['ifinput']

                    for s in ifinputs:
                        self.log.debug("evaluating ifinput %s" % s)
                        ifinput = s.replace("%", rootname)
                        self.log.debug("evaluating ifinput %s" % ifinput)
                        input_docs = parse_doc(path, ifinput, {})
                        for input_doc in input_docs:
                            self.log.debug(input_doc.key())
                            inputs.append(input_doc.key())

                    if len(input_docs) == 0:
                        create = False

                if args.has_key('ifnoinput'):
                    ifinput = args['ifnoinput'].replace("%", rootname)
                    input_docs = parse_doc(path, ifinput, {})

                    if len(input_docs) > 0:
                        create = False

                if args.has_key('except'):
                    try:
                        except_re = re.compile(args['except'])
                    except sre_constants.error as e:
                        raise Exception("""You passed 'except' value of %s.
Please pass a valid Python-style regular expression for
'except', NOT a glob-style matcher. Error message from
re.compile: %s""" % (args['except'], e))
                    if re.match(except_re, f):
                        print "skipping %s for %s as it matches except pattern %s" % (
                                f,
                                input_directive,
                                args['except']
                                )
                        create = False

                if create:
                    # Filters can either be included in the name...
                    doc = Document(self.artifact_class, f, filters)
                    doc.virtual = virtual
                    # ...or they may be listed explicitly.
                    if args.has_key('filters'):
                        doc.filters += args['filters']


                    # Here we are assuming that if we get a key with blank args
                    # this should not override a previous key. A key which does
                    # have args should override any previous key.
                    key = doc.key()
                    self.log.debug("creating doc %s for glob %s" % (key, glob_string))

                    if self.members.has_key(key):
                        self.log.debug("found existing key %s" % key)
                        doc = self.members[key]
                    else:
                        self.log.debug("no existing key %s" % key)

                    if len(args) > 0:
                        self.log.debug("args: %s" % args)
                        doc.args = args
                        doc.use_all_inputs = args.has_key('allinputs')
                        for i in inputs:
                            doc.add_input_key(i)

                    if not hasattr(doc, 'args'):
                        doc.args = args

                    doc.controller_args = self.args
                    self.members[key] = doc
                    docs.append(doc) # just a local list
            return docs

        def get_pos(member):
            key = member.key()
            return self.members.keys().index(key)

        def depend(parent, child):
            self.depends.append((get_pos(child), get_pos(parent)))

        self.members = OrderedDict()
        self.depends = []

        # Create Document objects for all docs.
        self.log.debug("About to process config:\n" + pprint.pformat(self.config))
        for path, config in self.config.iteritems():
            ### @export "features-global-args-1"
            if config.has_key("$globals"):
                global_args = config["$globals"]
            else:
                global_args = {}

            if self.args.globals:
                global_args.update(self.args.globals)

            for k, v in config.iteritems():
                local_args = global_args.copy()
                local_args.update(v)
                for kg in global_args.keys():
                    if local_args.has_key(kg):
                        if isinstance(local_args[kg], dict):
                            local_args[kg].update(global_args[kg])
                parse_doc(path, k, local_args)
            ### @end

        # Determine dependencies
        for doc in self.members.values():
            doc.finalize_inputs(self.members)
            for input_doc in doc.inputs:
                depend(doc, input_doc)

        ordering = topological_sort(range(len(self.members)), self.depends)

        self.db = None
        self.batch_id = None
        if dexy.db.DB_AVAILABLE:
            print "initializing database"
            self.db = dexy.db.Db()
            self.batch_id = self.db.next_batch_id()

        ordered_members = OrderedDict()
        for i in ordering:
            key = self.members.keys()[i]
            ordered_members[key] = self.members[key]
        self.members = ordered_members

    def run(self):
        return [doc.run(self) for doc in self.members.values()]

    def setup_and_run(self):
        if not hasattr(self, 'config_file'):
            self.config_file = '.dexy'
        self.register_filters()
        self.process_config()
        self.docs = self.run()
        return self.docs



from dexy.dexy_filter import DexyFilter
import os
import subprocess

class AsciidocHandler(DexyFilter):
    """IN DEVELOPMENT. Converts .txt files with asciidoc markup to HTML or
    XML."""
    VERSION = "asciidoc --version"
    EXECUTABLE = "asciidoc"
    INPUT_EXTENSIONS = [".txt"]
    OUTPUT_EXTENSIONS = [".html", ".xml"]
    ALIASES = ['asciidoc']
    FINAL = True

    def process(self):
        self.artifact.generate_workfile()
        workfile = self.artifact.work_filename()
        outfile = self.artifact.filename()

        extension = self.artifact.ext

        if extension == ".html":
            # TODO check here if we are on asciidoc 8.6.5,
            # any lower will throw an error if we try to use html5
            backend = "html5"
        elif extension == ".xml":
            backend = "docbook45"
        else:
            raise Exception("unexpected file extension in asciidoc handler %s" % extension)

        command = "%s -b %s -d book -o %s %s" % (self.EXECUTABLE, backend, outfile, workfile)
        self.log.debug(command)

        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None

        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.artifacts_dir,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)

        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

        self.handle_subprocess_proc_return(proc.returncode, stderr)




from BeautifulSoup import BeautifulSoup
from ansi2html import Ansi2HTMLConverter
from pynliner import Pynliner
import StringIO
import datetime
import gc
import logging
import os
import re

# http://code.activestate.com/recipes/361668/#c2
class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
    def __getattr__(self, name):
        return self[name]

def ansi_output_to_html(ansi_text):
    try:
        converter = Ansi2HTMLConverter()
        html = converter.convert(ansi_text)
    except IOError as e:
        if re.search("templates/header.mak", str(e)):
            print e
            raise Exception("Your installation of ansi2html is missing some template files, please try 'pip install --upgrade ansi2html' or install from source.")
        raise e

    log = logging.getLogger("pynliner")
    log.setLevel(logging.ERROR) # TODO Allow this to be changed.
    logstream = StringIO.StringIO()
    handler = logging.StreamHandler(logstream)
    log.addHandler(handler)
    try:
        p = Pynliner(log)
    except TypeError:
        print "========== Start of harmless but annoying CSS errors..."
        print "You can install pynliner from source (https://github.com/rennat/pynliner.git) or version > 0.2.1 to get rid of these"
        p = Pynliner()

    p.from_string(html)
    html_with_css_inline = p.run()

    # Ansi2HTMLConverter returns a complete HTML document, we just want body
    doc = BeautifulSoup(html_with_css_inline)
    return doc.body.renderContents()

def print_string_diff(str1, str2):
    msg = ""
    for i, c1 in enumerate(str1):
        if len(str2) > i:
            c2 = str2[i]
            if c1 == c2:
                flag = ""
            else:
                flag = " <---"
            if ord(c1) > ord('a') and ord(c2) > ord('a'):
                msg = msg + "\n%5d: %s\t%s\t\t%s\t%s %s" % (i, c1, c2,
                                              ord(c1), ord(c2), flag)
            else:
                msg = msg + "\n%5d:  \t \t\t%s\t%s %s" % (i, ord(c1),
                                              ord(c2), flag)
        else:
            flag = "<---"
            msg = msg + "\n%5d:  \t \t\t%s %s" % (i, ord(c1), flag)
    return msg

# Based on http://nedbatchelder.com/code/utilities/wh.py
def command_exists(cmd_name):
    path = os.environ["PATH"]
    if ";" in path:
        path = filter(None, path.split(";"))
    else:
        path = filter(None, path.split(":"))

    if os.environ.has_key("PATHEXT"):
        # Present on windows, returns e.g. '.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC'
        pathext = os.environ["PATHEXT"]
        pathext = filter(None, pathext.split(";"))
    else:
        # Not windows, look for exact command name.
        pathext = [""]

    cmd_found = False
    for d in path:
        for e in pathext:
            filepath = os.path.join(d, cmd_name + e)
            if os.path.exists(filepath):
                cmd_found = True
                break
    return cmd_found

#def profile_memory(description=None):
#    file_name = "memory.log"
#    new_file = not os.path.exists(file_name)
#    f = open("memory.log", "a")
#    if new_file:
#        f.write("vm_size\tgc_count\ttimestamp\tdescription\n")
#    pid = os.getpid()
#    vm_size = MemoryMonitor.getVmSize(pid)
#    gc_count = len(gc.get_objects())
#    timestamp = datetime.datetime.now()
#    f.write("%s\t%s\t%s\t%s\n" % (vm_size, gc_count, timestamp, description))
#    f.close()



from dexy.dexy_filter import DexyFilter
import os
import subprocess

class ProcessStdoutHandler(DexyFilter):
    """
    Runs python script and returns STDOUT.
    """
    EXECUTABLE = 'python'
    VERSION = 'python --version'
    WINDOWS_EXECUTABLE = 'python'
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['py', 'python', 'pyout']

    def process(self):
        # allow running code from different directory
        cwd = self.artifact.artifacts_dir
        if self.artifact.args.has_key('cwd'):
            cwd = os.path.join(cwd, self.artifact.args['cwd'])
            print "running from", os.path.abspath(cwd)

        self.artifact.generate_workfile()
        wf = os.path.abspath(os.path.join(self.artifact.artifacts_dir,
                                          self.artifact.work_filename()))

        if hasattr(self, 'instance_executable'):
            command = "%s %s" % (self.instance_executable(), wf)
        else:
            command = "%s %s" % (self.executable(), wf)

        cla = self.artifact.command_line_args()
        if cla:
            command = "%s %s" % (command, cla)

        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None

        self.artifact.log.debug(command)

        proc = subprocess.Popen(command, shell=True,
                                cwd=cwd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=env)
        stdout, stderr = proc.communicate()
        self.artifact.data_dict['1'] = stdout
        self.artifact.stdout = stderr

        self.handle_subprocess_proc_return(proc.returncode, stderr)

class BashHandler(ProcessStdoutHandler):
    """
    Runs bash script and returns STDOUT.
    """
    EXECUTABLE = 'bash'
    VERSION = 'bash --version'
    INPUT_EXTENSIONS = [".sh", ".bash", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['bash']

class PhpHandler(ProcessStdoutHandler):
    """
    Runs php script and returns STDOUT.
    """
    EXECUTABLE = 'php'
    VERSION = 'php --version'
    INPUT_EXTENSIONS = [".php"]
    OUTPUT_EXTENSIONS = [".html", ".txt"]
    ALIASES = ['php']

class EscriptHandler(ProcessStdoutHandler):
    """
    Runs escript (erlang) and returns STDOUT.
    """
    EXECUTABLE = 'escript'
    INPUT_EXTENSIONS = [".erl"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['escript']

class LuaHandler(ProcessStdoutHandler):
    """
    Runs lua script and returns STDOUT.
    """
    EXECUTABLE = 'lua'
    VERSION = 'lua -v'
    INPUT_EXTENSIONS = ['.lua']
    OUTPUT_EXTENSIONS = ['.txt']
    ALIASES = ['lua']

class RedclothHandler(ProcessStdoutHandler):
    """
    Runs redcloth, converts textile markup to HTML.
    """
    EXECUTABLE = 'redcloth'
    VERSION = 'redcloth --version'
    INPUT_EXTENSIONS = [".txt", ".textile"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['redcloth', 'textile']

class RedclothLatexHandler(ProcessStdoutHandler):
    """
    Runs redcloth, converts textile markup to LaTeX.
    """
    EXECUTABLE = 'redcloth -o latex'
    VERSION = 'redcloth --version'
    INPUT_EXTENSIONS = [".txt", ".textile"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['redclothl', 'latextile']

class Rst2HtmlHandler(ProcessStdoutHandler):
    EXECUTABLE = 'rst2html.py'
    VERSION = 'rst2html.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['rst2html']
    FINAL = True

class Rst2LatexHandler(ProcessStdoutHandler):
    EXECUTABLE = 'rst2latex.py'
    VERSION = 'rst2latex.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['rst2latex']

class Rst2BeamerHandler(ProcessStdoutHandler):
    EXECUTABLE = 'rst2beamer'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['rst2beamer']

class SloccountHandler(ProcessStdoutHandler):
    EXECUTABLE = 'sloccount'
    VERSION = 'sloccount --version'
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['sloc', 'sloccount']

class RubyStdoutHandler(ProcessStdoutHandler):
    EXECUTABLE = 'ruby'
    VERSION = 'ruby --version'
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['rb']

class RdConvHandler(ProcessStdoutHandler):
    """Convert R documentation to other formats."""
    EXECUTABLE = "R CMD Rdconv"
    VERSION = "R CMD Rdconv -v"
    INPUT_EXTENSIONS = ['.Rd']
    OUTPUT_EXTENSIONS = ['.txt', '.html', '.tex', '.R']
    ALIASES = ['rdconv']
    EXTENSION_TO_FORMAT = {
        '.txt' : 'txt',
        '.html' : 'html',
        '.tex' : 'latex',
        '.R' : 'example'
    }

    def instance_executable(self):
        return "%s --type=%s" % (self.EXECUTABLE, self.EXTENSION_TO_FORMAT[self.artifact.ext])

class RagelRubyDotHandler(ProcessStdoutHandler):
    """
    Generates state chart in .dot format of ragel state machine for ruby.
    """
    INPUT_EXTENSIONS = [".rl"]
    OUTPUT_EXTENSIONS = [".dot"]
    ALIASES = ['rlrbd', 'ragelrubydot']
    VERSION = 'ragel --version'
    EXECUTABLE = 'ragel -R -V'




from dexy.artifact import Artifact
from dexy.controller import Controller
from ordereddict import OrderedDict
import json
import os

class FileSystemJsonArtifact(Artifact):
    """Artifact which persists data by writing to the file system (default type
    of Artifact)"""

    FILTERS = Controller().find_filters()

    # Metadata
    def meta_filename(self):
        return "%s-meta.json" % (self.hashstring)

    def meta_filepath(self):
        return os.path.join(self.artifacts_dir, self.meta_filename())

    def save_meta(self):
        m = {}
        attrs_to_persist = set(self.META_ATTRS + self.HASH_WHITELIST) - set(['input_data_dict', 'inputs'])
        for a in attrs_to_persist:
            if hasattr(self, a):
                v = getattr(self, a)
                m[a] = v

        m['inputs'] = {}
        for k, a in self.inputs().iteritems():
            a.save()
            m['inputs'][k] = a.hashstring

        f = open(self.meta_filepath(), "w")
        try:
            json.dump(m, f)
        except UnicodeDecodeError as e:
            print e
            print m
            raise Exception("Binary data present in %s" % self.key)
        f.close()

    def load_meta(self):
        f = open(self.meta_filepath(), "r")
        m = json.load(f)
        f.close()

        self._inputs = dict((k, self.__class__.retrieve(h)) for (k, h) in m.pop('inputs').iteritems())

        for k, v in m.iteritems():
            setattr(self, k, v)

        # We only store filter name, not filter class, need to retrieve class from name
        if hasattr(self, "filter_name") and not hasattr(self, "filter_class"):
            self.filter_class = [k for n,k in self.FILTERS.iteritems() if k.__name__ == self.filter_name][0]

    # Input
    def load_input(self):
        """Load input data into memory, if applicable."""
        if self.binary_input:
            #not loading non-binary input
            pass
        elif self.initial:
            #initial artifact has no input
            pass
        elif self.additional:
            #additional artifact has no input
            pass
        elif len(self.input_data_dict) > 0:
            #we already have input data in memory
            pass
        elif not hasattr(self, 'previous_cached_output_filepath'):
            #no previous cached output, can't load
            pass
        else:
            f = open(self.previous_cached_output_filepath, "r")
            data_dict = json.load(f)
            f.close()

            self.input_data_dict = OrderedDict() # maybe unnecessary
            for x in sorted(data_dict.keys()):
                k = x.split(":", 1)[1]
                self.input_data_dict[k] = data_dict[x]

    # Output
    def cached_output_filename(self):
        return "%s-output.json" % (self.hashstring)

    def cached_output_filepath(self):
        return os.path.join(self.artifacts_dir, self.cached_output_filename())

    def is_output_cached(self):
        # TODO add checksums to verify data hasn't changed
        if self.binary_output:
            return self.is_canonical_output_cached()
        else:
            return self.is_json_output_cached() and self.is_canonical_output_cached()

    def is_json_output_cached(self):
        fp = self.cached_output_filepath()
        return os.path.isfile(fp) and (os.path.getsize(fp) > 0)

    def is_canonical_output_cached(self):
        fp = self.filepath()
        return os.path.isfile(fp) and (os.path.getsize(fp) > 0)

    def save_output(self):
        if not self.is_complete():
            raise Exception("should not be calling save_output unless artifact is complete")

#        if self.is_output_cached():
#            print "we will be overwriting existing files"

        if not self.binary_output:
            if not self.data_dict or len(self.data_dict) == 0:
                # Our filter has written directly to an output file
                # We need to load this into memory first
                self.data_dict = OrderedDict()
                f = open(self.filepath(), 'r')
                data = f.read()
                f.close()
                self.data_dict['1'] = data

            # need to preserve ordering but we can't serialize OrderedDict
            # using JSON, so add sortable numbers to keys to preserve order
            data_dict = {}
            MAX = 10000
            if len(self.data_dict) >= MAX:
                raise Exception("""There is an arbitrary limit of %s dict items,
                               you can increase this if you need to.""" % MAX)
            i = -1
            for k, v in self.data_dict.iteritems():
                i += 1
                data_dict["%04d:%s" % (i, k)] = v

            # Write the JSON file.
            f = open(self.cached_output_filepath(), "w")
            try:
                json.dump(data_dict, f)
            except UnicodeDecodeError as e:
                print e
                print self.binary_output
                raise Exception(self.key)

            f.close()

            # Write the canonical file.
            f = open(self.filepath(), 'w')
            f.write(self.output_text())
            f.close()

    def load_output(self):
        if not self.is_complete():
            raise Exception("should not be calling load_output unless artifact is complete")

        if not self.binary_output:
            f = open(self.cached_output_filepath(), "r")
            data_dict = json.load(f)
            f.close()

            self.data_dict = OrderedDict() # maybe unnecessary
            for x in sorted(data_dict.keys()):
                k = x.split(":", 1)[1]
                self.data_dict[k] = data_dict[x]




from dexy.version import VERSION
from dexy.utils import AttrDict
from ordereddict import OrderedDict
import glob
import hashlib
import inspect
import logging
import os
import shutil
import sys
import time
import traceback

class Artifact(object):
    META_ATTRS = [
        'binary_input',
        'binary_output',
        'key',
        'document_key',
        'name',
        'ext',
        'stdout',
        'final',
        'additional',
        'initial',
        'state',
        'batch_id',
        'output_hash',
        'elapsed'
    ]

    HASH_WHITELIST = [
        'args',
        'artifact_class_source',
        'dexy_version',
        'dirty',
        'dirty_string',
        'ext',
        'filter_name',
        'filter_source',
        'filter_version',
        'next_filter_name',
        'inputs',
        'input_data_dict',
        'input_ext',
        'key'
    ]

    BINARY_EXTENSIONS = [
        '.gif',
        '.jpg',
        '.png',
        '.pdf',
        '.zip',
        '.tgz',
        '.gz',
        '.eot',
        '.ttf',
        '.woff',
        '.swf'
    ]

    def __init__(self):
        if not hasattr(self.__class__, 'SOURCE_CODE'):
            self.__class__.SOURCE_CODE = inspect.getsource(self.__class__)

        self._inputs = {}
        self.additional = None

        # In actual usage, these initialized values are replaced by optparse
        # Namespace or the argparse equivalent. These initialized values are
        # used in testing, AttrDict is used so we can simulate the globals
        # attr.
        self.args = AttrDict({})
        self.args.globals = {}

        self.artifact_class_source = self.__class__.SOURCE_CODE # TODO need artifact.py and also subclass source.
        self.artifacts_dir = 'artifacts' # TODO don't hard code
        self.binary_input = None
        self.binary_output = None
        self.data_dict = OrderedDict()
        self.dexy_version = VERSION
        self.dirty = False
        self.document_key = None
        self.final = None
        self.initial = None
        self.input_data_dict = OrderedDict()
        self.key = None
        self.log = logging.getLogger()
        self.state = 'new'

    def is_complete(self):
        return str(self.state) == 'complete'

    @classmethod
    def retrieve(klass, hashstring):
        artifact = klass()
        artifact.hashstring = hashstring
        artifact.load()
        return artifact

    def load(self):
        self.load_meta()
        self.load_input()
        if self.is_complete():
            self.load_output()

    def load_inputs(self):
        for a in self.inputs():
            a.load()

    def save(self):
        if self.is_abstract():
            pass # For testing.
        elif not self.hashstring:
            raise Exception("can't persist an artifact without a hashstring!")
        else:
            self.save_meta()
            if self.is_complete() and not self.is_output_cached():
                self.save_output()

    def is_abstract(self):
        return not hasattr(self, 'save_meta')

    def setup_from_filter_class(self, filter_class):
        if not hasattr(filter_class, 'SOURCE_CODE'):
            filter_class.SOURCE_CODE = inspect.getsource(filter_class)

        if self.final is None and filter_class.FINAL is not None:
            self.final = filter_class.FINAL

        self.filter_class = filter_class
        self.filter_name = filter_class.__name__
        self.filter_source = filter_class.SOURCE_CODE
        self.filter_version = filter_class.version(self.log)

    def setup_from_previous_artifact(self, previous_artifact):
        assert self.filter_class
        if self.final is None:
            self.final = previous_artifact.final
        self.binary_input = previous_artifact.binary_output
        self.input_ext = previous_artifact.ext
        self.input_data_dict = previous_artifact.data_dict
        self.document_key = previous_artifact.document_key
        self.controller_args = previous_artifact.controller_args

        # The 'canonical' output of previous artifact
        self.previous_artifact_filename = previous_artifact.filename()
        self.previous_artifact_filepath = previous_artifact.filepath()
        self.previous_canonical_filename = previous_artifact.canonical_filename()
        # The JSON output of previous artifact
        if not previous_artifact.binary_output:
            self.previous_cached_output_filepath = previous_artifact.cached_output_filepath()

        self._inputs.update(previous_artifact.inputs())
        # Need to loop over each artifact's inputs in case extra ones have been
        # added anywhere.
        for k, a in previous_artifact.inputs().iteritems():
            self._inputs.update(a.inputs())

        if hasattr(self, 'next_filter_class'):
            next_inputs = self.next_filter_class.INPUT_EXTENSIONS
        else:
            next_inputs = None

        self.ext = self.filter_class.output_file_extension(
                previous_artifact.ext,
                self.name,
                next_inputs
                )
        self.binary_output = self.filter_class.BINARY
        self.state = 'setup'

    @classmethod
    def setup(klass, doc, artifact_key, filter_class = None, previous_artifact = None):
        """Set up a new artifact."""
        artifact = klass()

        artifact.args = doc.args
        artifact.dexy_args = doc.controller.args
        artifact.artifacts_dir = doc.controller.artifacts_dir
        artifact.key = artifact_key
        artifact.log = doc.log
        artifact.name = doc.name

        if artifact.args.has_key('final'):
            # TODO move this into initial setup? Should only be needed once...
            artifact.final = artifact.args['final']
        elif hasattr(previous_artifact, 'final') and previous_artifact.final is not None:
            artifact.final = previous_artifact.final

        if filter_class:
            artifact.setup_from_filter_class(filter_class)

        if doc.next_filter_class():
            artifact.next_filter_name = doc.next_filter_class().__name__
            artifact.next_filter_class = doc.next_filter_class()

        if previous_artifact:
            artifact.setup_from_previous_artifact(previous_artifact)
        else:
            # This is an initial artifact
            artifact.initial = True
            artifact._inputs = doc.input_artifacts()
            artifact.ext = os.path.splitext(doc.name)[1]
            artifact.binary_input = (doc.ext in artifact.BINARY_EXTENSIONS)
            artifact.set_data(doc.initial_artifact_data())
            if not artifact.data_dict:
                raise Exception("no data dict!")
            elif len(artifact.data_dict) == 0:
                raise Exception("data dict has len 0!")

            if os.path.basename(doc.name).startswith("_"):
                artifact.final = False
            artifact.state = 'complete'

        if artifact.binary_output is None:
            artifact.set_binary_from_ext()

        artifact.set_hashstring()

        if doc.controller.db and False:
            # disabled temporarily since it's very slow...
            doc.controller.db.insert_artifact(artifact, doc.controller.batch_id)
        return artifact

    def run(self):
        start_time = time.time()

        if not self.is_complete():
            # We have to actually run things...
            if not self.filter_class:
                classes = [k for k in self.controller.handlers.values() if k.__name__ == self.filter_name]
                if len(classes) == 0:
                    raise Exception("no filter class %s found" % self.filter_name)
                self.filter_class = classes[0]

            # Set up instance of filter.
            filter_instance = self.filter_class()
            filter_instance.artifact = self
            filter_instance.log = self.log

            try:
                filter_instance.process()
            except Exception as e:
                print "Error occurred while running", self.key
                x, y, tb = sys.exc_info()
                print "Original traceback:"
                traceback.print_tb(tb)
                pattern = os.path.join(self.artifacts_dir, self.hashstring)
                files_matching = glob.glob(pattern)
                if len(files_matching) > 0:
                    print "Here are working files which might have clues about this error:"
                    for f in files_matching:
                        print f
                raise e

            h = hashlib.sha512()

            if self.data_dict and len(self.data_dict) > 0:
                h.update(self.output_text())

            elif self.is_canonical_output_cached:
                self.state = 'complete'
                self.save()

                f = open(self.filepath(), "rb")
                while True:
                    data = f.read(h.block_size)
                    if not data:
                        break
                    h.update(data)

            else:
                raise Exception("data neither in memory nor on disk")

            self.output_hash = h.hexdigest()

            self.state = 'complete'
            finish_time = time.time()
            self.elapsed = finish_time - start_time
            self.save()
        else:
            self.log.debug("using cached art %s" % self.key)


    def add_additional_artifact(self, key_with_ext, ext):
        """create an 'additional' artifact with random hashstring"""
        new_artifact = self.__class__()
        new_artifact.key = key_with_ext
        new_artifact.ext = ".%s" % ext
        new_artifact.final = True
        new_artifact.additional = True
        new_artifact.set_binary_from_ext()
        new_artifact.artifacts_dir = self.artifacts_dir
        new_artifact.set_hashstring()
        self.add_input(key_with_ext, new_artifact)
        return new_artifact

    def add_input(self, key, artifact):
        self._inputs[key] = artifact

    def inputs(self):
        return self._inputs

    def set_binary_from_ext(self):
        # TODO list more binary extensions or find better way to do this
        if self.ext in self.BINARY_EXTENSIONS:
            self.binary_output = True
        else:
            self.binary_output = False

    def set_data(self, data):
        self.data_dict['1'] = data

    def set_data_from_artifact(self):
        f = open(self.filepath())
        self.data_dict['1'] = f.read()

    def is_loaded(self):
        return hasattr(self, 'data_dict') and len(self.data_dict) > 0

    def hash_dict(self):
        """Return the elements for calculating the hashstring."""
        if self.dirty:
            self.dirty_string = time.gmtime()

        hash_dict = OrderedDict()

        hash_dict['inputs'] = OrderedDict()
        sorted_input_keys = sorted(self.inputs().keys())
        for k in sorted_input_keys:
            hash_dict['inputs'][str(k)] = str(self.inputs()[k].hashstring)

        for k in self.HASH_WHITELIST:
            if self.__dict__.has_key(k):
                v = self.__dict__[k]
                if hasattr(v, 'items'):
                    hash_v = OrderedDict()
                    for k1, v1 in v.items():
                        hash_v[str(k1)] = hashlib.md5(str(v1)).hexdigest()
                else:
                    hash_v = hashlib.md5(str(v)).hexdigest()
                hash_dict[str(k)] = hash_v

        return hash_dict

    def set_hashstring(self):
        hash_data = str(self.hash_dict())
        self.hashstring = hashlib.md5(hash_data).hexdigest()

        if False:
            for k, v in self.hash_dict().iteritems():
                if hasattr(v, 'items'):
                    print k, ":"
                    for k1, v1 in v.items():
                        print "  ", k1, ":", v1
                else:
                    print k, ":", v
            print hash_data
            print ">>>>>", self.hashstring

        try:
            self.load()
        except AttributeError as e:
            if not self.is_abstract():
                raise e
        except IOError as e:
            self.save_meta()

    def command_line_args(self):
        """
        Allow specifying command line arguments which are passed to the filter
        with the given key. Note that this does not currently allow
        differentiating between 2 calls to the same filter in a single document.
        """
        if self.args.has_key('args'):
            args = self.args['args']
            last_key = self.key.rpartition("|")[-1]
            if args.has_key(last_key):
                return args[last_key]

    def input_text(self):
        return "".join([v for k, v in self.input_data_dict.items()])

    def output_text(self):
        return "".join([v for k, v in self.data_dict.items()])

    def use_canonical_filename(self):
        """Returns the canonical filename after saving contents under this name
        in the artifacts directory."""
        self.write_to_file(os.path.join(self.artifacts_dir,
                                        self.canonical_filename()))
        return self.canonical_filename()

    def write_to_file(self, filename):
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname) and not dirname == '':
            os.makedirs(dirname)
        shutil.copyfile(self.filepath(), filename)

    def work_filename(self):
        return "%s.work%s" % (self.hashstring, self.input_ext)

    def generate_workfile(self, work_filename = None):
        if not work_filename:
            work_filename = self.work_filename()
        work_path = os.path.join(self.artifacts_dir, work_filename)
        work_file = open(work_path, "w")
        work_file.write(self.input_text())
        work_file.close()

    def temp_filename(self, ext):
        return "%s.work%s" % (self.hashstring, ext)

    def open_tempfile(self, ext):
        tempfile_path = os.path.join(self.artifacts_dir, self.temp_filename(ext))
        open(tempfile_path, "w")

    def temp_dir(self):
        return os.path.join(self.artifacts_dir, self.hashstring)

    def create_temp_dir(self):
        shutil.rmtree(self.temp_dir(), ignore_errors=True)
        os.mkdir(self.temp_dir())

    def canonical_basename(self):
        return os.path.basename(self.canonical_filename())

    def canonical_filename(self):
        fn = os.path.splitext(self.key.split("|")[0])[0]
        return "%s%s" % (fn, self.ext)

    def long_canonical_filename(self):
        return "%s%s" % (self.key.replace("|", "-"), self.ext)

    def filename(self):
        if not hasattr(self, 'ext'):
            raise Exception("artifact %s has no ext" % self.key)
        return "%s%s" % (self.hashstring, self.ext)

    def filepath(self):
        return os.path.join(self.artifacts_dir, self.filename())




try:
    import git
    USE_GIT = True
except ImportError:
    USE_GIT = False

import StringIO
import hashlib
import json
import logging
import os
import urllib2


class Document(object):
    def __init__(self, artifact_class, name_or_key, filters = []):
        self.artifact_class = artifact_class
        self.name = name_or_key.split("|")[0]
        self.ext = os.path.splitext(self.name)[1]
        self.filters = name_or_key.split("|")[1:]
        self.filters += filters
        self.inputs = []
        self.input_keys = []
        self.artifacts = []

        # Set up document log.
        self.logstream = StringIO.StringIO()
        self.log = logging.getLogger(self.key())
        self.log.setLevel(logging.DEBUG) # TODO Allow this to be changed.
        self.log.propagate = 0 # Stops logs being written to STDOUT if another library redirects root logger.
        handler = logging.StreamHandler(self.logstream)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        self.log.addHandler(handler)

        # Log all document log messages to main dexy log.
        try:
            self.log.addHandler(logging.getLogger('dexy').handlers[0])
        except IndexError:
            pass

        self.use_all_inputs = False

    def final_artifact(self):
        return self.last_artifact

    def key(self):
        return "%s|%s" % (self.name, "|".join(self.filters))

    def add_input_key(self, input_key):
        if not input_key in self.input_keys:
            self.input_keys.append(input_key)

    def finalize_inputs(self, members_dict):
        if self.use_all_inputs:
            for doc in members_dict.values():
                also_all_inputs = doc.use_all_inputs
                specified = doc.key() in self.input_keys

                rel1 = os.path.dirname(doc.name)
                rel2 = os.path.dirname(self.name)
                if rel1 == '':
                    rel1 = "."
                if rel2 == "":
                    rel2 = "."

                in_subdir = (".." in os.path.relpath(rel2, rel1))

                if specified or in_subdir or not also_all_inputs:
                    self.inputs.append(doc)
        else:
            self.inputs = [members_dict[k] for k in self.input_keys]

    def next_filter_alias(self):
        if self.at_last_step():
            return None
        else:
            return self.filters[self.step]

    def next_filter_class(self):
        alias = self.next_filter_alias()
        if alias:
            return self.controller.get_handler_for_alias(alias)

    def at_last_step(self):
        return (len(self.filters) == self.step)

    def input_artifacts(self):
        input_artifacts = {}
        for i in self.inputs:
            k = i.key()
            a = i.final_artifact()
            if not a:
                raise Exception("No final artifact for document %s" % i.key())
            input_artifacts[k] = a
        return input_artifacts

    def initial_artifact_data(self):
        if self.args.has_key('url'):
            url = self.args['url']
            filename = os.path.join(self.controller.artifacts_dir, self.name)
            header_filename = "%s.headers" % filename

            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))

            header_dict = {}
            if os.path.exists(header_filename):
                header_file = open(header_filename, "r")
                header_dict = json.load(header_file)
                header_file.close()

            request = urllib2.Request(url)

            # TODO add an md5 of the file to the header dict so we can check
            # that the etag/last-modified header is the corresponding one
            # TODO invalidate the hash if URL has changed
            # TODO get modification time of file via ftplib
            # TODO implement handlers for version control systems
            # TODO use FileHandler for default case?

            # Add any custom headers...
            # note that headers are ignored by anything other than http
            if header_dict.has_key('ETag') and os.path.exists(filename):
                request.add_header('If-None-Match', header_dict['ETag'])
            elif header_dict.has_key('Last-Modified') and os.path.exists(filename):
                request.add_header('If-Modifed-Since', header_dict['Last-Modified'])

            if self.controller.use_local_files and os.path.exists(filename):
                f = open(filename, "r")
                data = f.read()
                f.close()
            else:
                if self.controller.use_local_files:
                    print "local file %s not found, fetching remote url" % filename

                try:
                    u = urllib2.urlopen(request)
                    print "downloading contents of %s" % url
                    url_contents = u.read()

                    # Save the contents in our local cache
                    f = open(filename, "wb")
                    f.write(url_contents)
                    f.close()

                    # Save header info in our local cache
                    header_dict = {}
                    for s in u.info().headers:
                        a = s.partition(":")
                        header_dict[a[0]] = a[2].strip()
                    json.dump(header_dict, open(header_filename, "w"))

                    data = url_contents
                except urllib2.URLError as err:
                    if os.path.exists(filename):
                        print "unable to fetch remote url %s because %s\nusing contents of %s" % (url, err, filename)
                    else:
                        raise Exception("unable to fetch remote url %s because %s\nno cache found in %s" % (url, err, filename))

                    f = open(filename, "r")
                    data = f.read()
                    f.close()
                except urllib2.HTTPError as err:
                    if err.code == 304:
                        print "received http status code %s, using contents of %s" % (err.code, filename)
                        f = open(filename, "r")
                        data = f.read()
                        f.close()
                    elif err.code == 404:
                        raise Exception("""received http status code %s while trying to fetch %s for %s""" % 
                                        (err.code, url, self.name))
                    else:
                        # Some other http error, we want to know about it.
                        print url
                        raise err

        elif self.args.has_key('repo') and self.args.has_key('path'):
            if not USE_GIT:
                raise Exception("you can't use repo/path unless you install GitPython")
            repo_url = self.args['repo']
            digest = hashlib.md5(repo_url).hexdigest()
            local_repo_dir = os.path.join(self.controller.artifacts_dir, "repository-%s" % digest)

            if os.path.exists(local_repo_dir):
                repo = git.Repo(local_repo_dir)
                o = repo.remotes.origin
                assert o.url == repo_url, "local repo exists but url %s does not match requested url %s" % (o.url, repo_url)
                o.pull() # TODO remember last pulled time so that don't do this too often? what if network unavailable?
            else:
                repo = git.Repo.clone_from(repo_url, local_repo_dir)

            # TODO be able to specify different commits/revisions
            tree = repo.heads.master.commit.tree
            blob = tree[self.args['path']]
            data = blob.data_stream.read()

        elif self.args.has_key('contents'):
            data = self.args['contents']

        elif self.virtual:
            data = None

        else:
            # Normal local file, just read the contents.
            f = open(self.name, "r")
            data = f.read()
            f.close()

        return data

    def create_initial_artifact(self):
        # Create and set up the new artifact.
        artifact = self.artifact_class.setup(self, self.name)
        artifact.elapsed = 0
        artifact.document_key = self.key()
        artifact.controller_args = self.controller_args

        # Add the new artifact to the document's list of artifacts.
        self.artifacts.append(artifact)

        return artifact

    def run(self, controller):
        self.controller = controller
        self.step = 0

        artifact = self.create_initial_artifact()
        artifact_key = artifact.key
        self.log.info("(step %s) [run] %s -> %s" % \
                 (self.step, artifact_key, artifact.filename()))
#        profile_memory("document-%s-step-%s" % (self.key(), self.step))
        self.last_artifact = artifact
        for f in self.filters:
            previous_artifact = artifact
            artifact_key += "|%s" % f
            self.step += 1

            FilterClass = self.controller.get_handler_for_alias(f)
            artifact = self.artifact_class.setup(self, artifact_key, FilterClass, previous_artifact)

            self.log.info("(step %s) [run] %s -> %s" % \
                     (self.step, artifact_key, artifact.filename()))

            artifact.run()

            self.last_artifact = artifact
#      	    profile_memory("document-%s-step-%s" % (self.key(), self.step))


        # Make sure all additional inputs are saved.
        for k, a in artifact._inputs.iteritems():
            if not a.is_complete():
                a.state = 'complete'
                a.save()

#        profile_memory("document-%s-complete" % (self.key()))
        return self



# Collect per-process memory use over time.
# Written by Lucas C. Villa Real <lucasvr@us.ibm.com>
# Released under the GNU GPL version 2.

import os

def parseStatusFile(pid, field) :
	contents = ""
	fp = open("/proc/%d/status" %pid)
	while True :
		buf = fp.readline(4096)
		if not buf :
			break
		idx = buf.find(field)
		if idx >= 0 :
			contents = buf[idx:].split()[1]
			break
	fp.close()
	return contents 

def getVmSize(pid) :
	vmSize = parseStatusFile(pid, "VmSize:")
	if len(vmSize) :
		return int(vmSize)
	return -1

def getVmSwap(pid) :
	vmSwap = parseStatusFile(pid, "VmSwap:")
	if len(vmSwap) :
		return int(vmSwap)
	return -1

def getPPid(pid) :
	ppid = parseStatusFile(pid, "PPid:")
	if len(ppid) :
		return int(ppid)
	return -1

def getProcessName(pid) :
	fp = open("/proc/%d/cmdline" %pid)
	buf = fp.readline(1024)
	fp.close()

	tokens = buf.split('\x00')
	if tokens[0].find('python') >= 0 or tokens[0].find('/bin/sh') >= 0 :
		processName = tokens[1]
	elif len(tokens[0]) > 0 :
		processName = tokens[0]
	else :
		fp = open("/proc/%d/stat" %pid)
		buf = fp.readline(1024)
		fp.close()

		idx_start = buf.find("(")
		idx_end = buf.find(")")
		if idx_start >= 0 and idx_end >= 0 :
			processName = buf[idx_start+1:idx_end]
		else :
			print "unknown --> ",
			print processName
			print "tokens  --> ",
			print tokens
			return ""

	if processName.find("/") >= 0  :
		processName = os.path.basename(processName)
	if processName.find(":") >= 0 :
		processName = processName.split(":")[0]
	if processName.find("-") == 0 :
		processName = processName[1:]

	return processName

def getTotalMemory() :
	totalMemory = 0
	fp = open("/proc/meminfo")
	for line in fp.readlines() :
		if line.find("MemTotal:") == 0 :
			totalMemory = int(line.split()[1])
			break
	fp.close()
	return totalMemory








from dexy.dexy_filter import DexyFilter
from ordereddict import OrderedDict
from pygments import highlight
from pygments.formatters import get_formatter_for_filename
from pygments.lexers import get_lexer_for_filename
from pygments.lexers.agile import PythonConsoleLexer
from pygments.lexers.agile import RubyConsoleLexer
from pygments.lexers.special import TextLexer
from pygments.lexers.web import JavascriptLexer

class PygHandler(DexyFilter):
    """
    Apply Pygments syntax highlighting.
    """
    INPUT_EXTENSIONS = [".*"]
    IMAGE_OUTPUT_EXTENSIONS = ['.png', '.bmp', '.gif', '.jpg', '.svg']
    OUTPUT_EXTENSIONS = [".html", ".tex"] + IMAGE_OUTPUT_EXTENSIONS
    ALIASES = ['pyg', 'pygments']
    FINAL = False

    def process_dict(self, input_dict):
        ext = self.artifact.input_ext
        name = "input_text%s" % ext
        # List any file extensions which don't map neatly to lexers.
        if ext == '.pycon':
            lexer = PythonConsoleLexer()
        elif ext == '.rbcon':
            lexer = RubyConsoleLexer()
        elif ext in ('.json', '.dexy'):
            lexer = JavascriptLexer()
        elif ext == '.Rd':
            lexer = TextLexer()
        else:
            lexer = get_lexer_for_filename(name)

        if self.artifact.args.has_key('pygments'):
            pygments_args = self.artifact.args['pygments']
        else:
            pygments_args = {}

        formatter_args = {'lineanchors' : self.artifact.document_key }
        if pygments_args.has_key('formatter'):
            formatter_args.update(pygments_args['formatter'])

        formatter = get_formatter_for_filename(self.artifact.filename(),
                **formatter_args)

        if self.artifact.ext in self.IMAGE_OUTPUT_EXTENSIONS:
            self.artifact.binary_output = True
            f = open(self.artifact.filepath(), 'w')
            f.write(highlight(self.artifact.input_text(), lexer, formatter))
            f.close()
        else:
            output_dict = OrderedDict()
            for k, v in input_dict.items():
                # TODO figure out where these characters are coming from and don't hard-code this.
                v = str(v.replace(" \x08", "").replace(chr(13), ""))
                try:
                    output_dict[k] = str(highlight(v, lexer, formatter))
                except UnicodeEncodeError as e:
                    self.artifact.log.warn("error processing section %s of file %s" % (k, self.artifact.key))
                    raise e
            return output_dict



from dexy.reporter import Reporter
import datetime
import os
import shutil
import tarfile

class TarzipOutputReporter(Reporter):
    """Reporter which creates a .tgz of all output files."""
    OUTPUT_DIR = "logs"
    DEFAULT = False

    def run(self):
        # TODO craete output dir if it doesn't exist
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        report_filename = os.path.join(self.OUTPUT_DIR, "output-%s.tgz" % timestamp)
        tar = tarfile.open(report_filename, mode="w:gz")
        subdir_name = "dexy-output"

        for doc in self.controller.docs:
            artifact = doc.final_artifact()

            if artifact.final:
                arcname = os.path.join(subdir_name, artifact.canonical_filename())
                tar.add(artifact.filepath(), arcname=arcname)

            for k, a in artifact._inputs.items():
                if not a:
                    print "no artifact exists for key", k
                else:
                    if a.additional and a.final:
                        if not a.is_complete():
                            a.state = 'complete'
                            a.save()
                            arcname = os.path.join(subdir_name, a.canonical_filename())
                            tar.add(a.filepath(), arcname=arcname)

        tar.close()
        shutil.copyfile(report_filename, "output-latest.tgz")

class InSituReporter(Reporter):
    """This is the InSitu Reporter"""
    DEFAULT = False

    def run(self):
        for doc in self.controller.docs:
            artifact = doc.final_artifact()

            fn = artifact.canonical_filename()

            if artifact.final:
                if os.path.exists(fn):
                    print "InSituReporter not overwriting existing file", fn
                else:
                    artifact.write_to_file(fn)

            for k, a in artifact._inputs.items():
                if not a:
                    print "no artifact exists for key", k
                else:
                    if a.additional and a.final:
                        if not a.is_complete():
                            a.state = 'complete'
                            a.save()
                        fn = a.canonical_filename()
                        if not os.path.exists(fn):
                            a.write_to_file(fn)

class OutputReporter(Reporter):
    """This is the OutputReporter"""

    REPORTS_DIR = 'output'
    def run(self):
        shutil.rmtree(self.REPORTS_DIR, ignore_errors=True)
        for doc in self.controller.docs:
            artifact = doc.final_artifact()

            fn = artifact.canonical_filename()
            fp = os.path.join(self.REPORTS_DIR, fn)

            if artifact.final:
                artifact.write_to_file(fp)

            for k, a in artifact._inputs.items():
                if not a:
                    print "no artifact exists for key", k
                else:
                    if a.additional and a.final:
                        if not a.is_complete():
                            a.state = 'complete'
                            a.save()
                        fn = a.canonical_filename()
                        fp = os.path.join(self.REPORTS_DIR, fn)
                        a.write_to_file(fp)

class LongOutputReporter(Reporter):
    """This is the LongOutputReporter"""

    REPORTS_DIR = 'output-long'
    def run(self):
        shutil.rmtree(self.REPORTS_DIR, ignore_errors=True)
        for doc in self.controller.docs:
            artifact = doc.final_artifact()

            fn = artifact.long_canonical_filename()
            fp = os.path.join(self.REPORTS_DIR, fn)

            if not artifact.is_complete():
                artifact.state = 'complete'
                artifact.save()

            artifact.write_to_file(fp)

            for k, a in artifact._inputs.items():
                if not a:
                    print "no artifact exists for key", k
                else:
                    fn = a.canonical_filename()
                    fp = os.path.join(self.REPORTS_DIR, fn)
                    a.write_to_file(fp)



from dexy.artifact import Artifact
from dexy.controller import Controller
from dexy.version import VERSION
from inspect import isclass
from logging.handlers import RotatingFileHandler
import cProfile
import logging
import os
import re
import shutil
import sys
import urllib

# Directories with this name should be excluded anywhere
EXCLUDE_DIRS_ALL_LEVELS = ['.bzr', '.hg', '.git', '.svn']

# Directories with these names should be excluded only at the project root
EXCLUDE_DIRS_ROOT = ['ignore', 'filters']

EXCLUDE_HELP = """Specify directory names to exclude from processing by dexy.
Directories with these names will be skipped anywhere in your project.
The following patterns are automatically excluded anywhere in your project: %s.
The directories designated for artifacts and logs are automatically excluded,
and the following directory names are also excluded if they appear at the root
level of your project: %s.
Any directory with a file named .nodexy will be skipped, and subdirectories of this will also be skipped.
""" % (", ".join(EXCLUDE_DIRS_ALL_LEVELS), ", ".join(EXCLUDE_DIRS_ROOT))


def artifact_class(artifact_class_name, log):
    artifact_classes = {}

    d = os.path.join(os.path.dirname(__file__), 'artifacts') # yes 'artifacts' not artifacts_dir
    for f in os.listdir(d):
        if f.endswith(".py") and f not in ["base.py", "__init__.py"]:
            log.debug("Loading artifacts in %s" % os.path.join(d, f))
            basename = f.replace(".py", "")
            module = "dexy.artifacts.%s" % basename
            try:
                __import__(module)
            except ImportError as e:
                log.warn("artifact defined in %s are not available: %s" % (module, e))

            if not sys.modules.has_key(module):
                continue

            mod = sys.modules[module]

            for k in dir(mod):
                klass = mod.__dict__[k]
                if isclass(klass) and not (klass == Artifact) and issubclass(klass, Artifact):
                    if artifact_classes.has_key(k):
                        raise Exception("duplicate artifact class name %s called from %s in %s" % (k, f, d))
                    artifact_classes[klass.__name__] = klass

    if not artifact_class_name in artifact_classes.keys():
        raise Exception("artifact class %s not available, maybe check dexy.log for clues" % artifact_class_name)
    return artifact_classes[artifact_class_name]


def setup_option_parser():
    project_base = os.path.abspath(os.curdir)
    os.chdir(project_base)

    def add_option(parser, *args, **kwargs):
        if hasattr(parser, 'add_option'):
            parser.add_option(*args, **kwargs)
        else:
            parser.add_argument(*args, **kwargs)

    try:
        # These are argparse-specific
        import argparse
        option_parser = 'argparse'
        parser = argparse.ArgumentParser()


        class GlobalsArgparseAction(argparse.Action):
            def __call__(self, parser, namespace, values, option_string=None):
                value_dict = getattr(namespace, self.dest)
                if not value_dict:
                    value_dict = {}
                if not "=" in values:
                    raise Exception("args passed to %s should be in KEY=value format, no '=' found in %s" % (option_string, values))
                k, v = values.split("=")
                value_dict[k] = v
                setattr(namespace, self.dest, value_dict)

        parser.add_argument(
            '--global',
            help="""Use to set a global variable which will be available within dexy tasks,
            use KEY=value syntax and specify option multiple times if you need to specify more than 1 key value pair
            e.g. VERSION=0.0.1""",
            action=GlobalsArgparseAction,
            dest='globals'
        )

        parser.add_argument(
            'dir',
            help='directory of files to process with dexy',
            default='.', nargs='?'
        )

        parser.add_argument(
            '-x', '--exclude',
            help=EXCLUDE_HELP,
            nargs='+'
        )

        parser.add_argument(
            '--use-reporters', '--use-reporter',
            help="list reporters to be run",
            nargs='+'
        )

        parser.add_argument(
            '-v', '--version',
            action='version',
            version='%%(prog)s %s' % VERSION
        )

    except ImportError:
        # These are optparse-specific
        import optparse
        option_parser = 'optparse'
        parser = optparse.OptionParser(version="%%prog %s" % VERSION)
        parser.add_option(
            '-x', '--exclude',
            help=EXCLUDE_HELP + ' Separate multiple directories with commas.'
        )

        parser.add_option(
            "--use-reporters",
            help="list reporters to be used. separate multiple reproters with commas"
        )

    # Remaining options are the same for argparse and optparse
    add_option(parser,
        '-n', '--no-recurse',
        dest='recurse',
        default=True,
        action='store_false',
        help='do not recurse into subdirectories (default: recurse)'
    )

    add_option(parser,
        '--no-tests',
        dest='run_tests',
        default=True,
        action='store_false',
        help='do not run any dexy tests (using the test filter) - helpful for populating test content'
    )

    add_option(parser,
        '-u', '--utf8',
        default=False,
        action='store_true',
        help='switch encoding to UTF-8 (default: don\'t change encoding)'
    )

    add_option(parser,
        '-p', '--purge',
        default=False,
        action='store_true',
        help='purge all artifacts before running dexy'
    )

    add_option(parser,
        '--cleanup',
        default=False,
        action='store_true',
        help='delete all dexy-generated directories and files (does not run dexy)'
    )

    add_option(parser,
        '--filters',
        default=False,
        action='store_true',
        help='list all available filters (does not run dexy)'
    )

    add_option(parser,
        '--profile',
        default=False,
        action='store_true',
        help='run dexy with cProfile'
    )

    add_option(parser,
        '--reporters',
        default=False,
        action='store_true',
        help='list all available reporters (does not run dexy)'
    )

    add_option(parser,
        '--artifact-class',
        default='FileSystemJsonArtifact',
        help='name of artifact class to use (default: FileSystemJsonArtifact)'
    )

    add_option(parser,
        '-a', '--artifacts-dir',
        default='artifacts',
        help='location of artifacts directory (default: artifacts)'
    )

    add_option(parser,
        '-l', '--logs-dir',
        default='logs',
        help="""location of logs directory (default: logs)
               dexy will create a dexy.log file in this directory
               reporters may create reports in this directory"""
    )

    add_option(parser,
        '-c', '--cache-dir',
        default='cache',
        help='DEPRECATED - does not do anything anymore, will be removed soon'
    )

    add_option(parser,
        '--local',
        default=False,
        action='store_true',
        help='Use cached local copies of remote urls - faster but might not be up to date'
        )

    add_option(parser,
        '--ignore-errors',
        default=False,
        action='store_true',
        help="""Don't raise an error if scripts return nonzero exit status
               (depends on the filter being written to support this)"""
        )

    add_option(parser,
        '--setup',
        default=False,
        action='store_true',
        help='Create artifacts and logs directory and generic .dexy file'
        )

    add_option(parser,
        '--serve',
        default=False,
        action='store_true',
        help='Start a static server within the output directory, will be visible on port 8000. (similar to python -m SimpleHTTPServer)'
        )

    add_option(parser,
        '-g', '--config',
        default='.dexy',
        help='name of configuration file'
    )

    add_option(parser,
        '-d', '--dangerous',
        default=False,
        action='store_true',
        help='Allow running remote URLs which may execute dangerous code, use with care.'
     )

    add_option(parser,
        '--no-reports',
        default=False,
        action='store_true',
        help="""Don't run reports when finished running Dexy."""
     )

    if (option_parser == 'argparse'):
        args = parser.parse_args()
        dir_name = args.dir
        if args.exclude:
            additional_excludes = args.exclude
        else:
            additional_excludes = []

        if args.use_reporters:
            use_reporters = args.use_reporters
        else:
            use_reporters = None

    elif (option_parser == 'optparse'):
        (args, argv) = parser.parse_args()
        if len(argv) == 0:
            dir_name = '.'
        else:
            dir_name = argv[0]
        if args.exclude:
            additional_excludes = args.exclude.split(',')
        else:
            additional_excludes = []

        if args.use_reporters:
            use_reporters = args.use_reporters.split(',')
        else:
            use_reporters = None
    else:
        raise Exception("unexpected option_parser %s" % option_parser)

    args.run_dexy = True

    if args.utf8:
        if (sys.getdefaultencoding() == 'UTF-8'):
            print "encoding is already UTF-8"
        else:
            print "changing encoding from %s to UTF-8" % sys.getdefaultencoding()
            reload(sys)
            sys.setdefaultencoding("UTF-8")

    if not os.path.exists(dir_name):
        raise Exception("file %s not found!" % dir_name)

    if args.setup:
        if not os.path.exists(args.artifacts_dir):
            os.mkdir(args.artifacts_dir)
        if not os.path.exists(args.logs_dir):
            os.mkdir(args.logs_dir)
        if not os.path.exists(".dexy"):
            f = open(".dexy", "w")
            f.write("{\n}\n")
            f.close()

    if args.config.startswith("http"):
        print "fetching remote config file %s" % args.config
        filename = os.path.basename(args.config)
        print "writing to %s" % filename
        f = urllib.urlopen(args.config)
        config_file = open(filename, "w")
        config_file.write(f.read())
        config_file.close()
        f.close()
        args.config = filename


    if not os.path.exists(args.artifacts_dir):
        path_to_artifacts_dir = os.path.join(project_base, args.artifacts_dir)
        raise Exception(
            """artifacts directory not found at '%s',
            please call dexy with --setup if you want to use
            this location as a dexy project root""" % (args.artifacts_dir)
        )

    if not os.path.exists(args.logs_dir):
        path_to_logs_dir = os.path.join(project_base, args.logs_dir)
        raise Exception(
            """logs directory not found at %s,
            please call dexy with --setup if you want to use
            this location as a dexy project root""" % (path_to_logs_dir)
        )

    # Set up main dexy log
    dexy_log = logging.getLogger("dexy")
    dexy_log.setLevel(logging.DEBUG)

    logfile = os.path.join(args.logs_dir, 'dexy.log')
    handler = RotatingFileHandler(logfile)
    dexy_log.addHandler(handler)
    if args.setup:
        # This might be the first time someone has run Dexy,
        # let them know where the logfile is
        print "dexy will log debugging information to", logfile

    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    args.artifact_class = artifact_class(args.artifact_class, dexy_log)

    if args.purge:
        dexy_log.warn("purging contents of %s" % args.artifacts_dir)
        shutil.rmtree(args.artifacts_dir)
        os.mkdir(args.artifacts_dir)

        # Each artifact class may need to do its own purging also
        if hasattr(args.artifact_class, 'purge'):
            args.artifact_class.purge()

    if args.cleanup:
        args.run_dexy = False
        print "purging contents of %s" % args.artifacts_dir
        shutil.rmtree(args.artifacts_dir)

        if os.path.exists(args.cache_dir):
            print "purging contents of %s" % args.cache_dir
            shutil.rmtree(args.cache_dir)

        if os.path.exists(args.logs_dir):
            print "purging contents of %s" % args.logs_dir
            shutil.rmtree(args.logs_dir)

        if os.path.exists('testdb') and False:
            os.remove('testdb')

        if os.path.exists('output-latest.tgz'):
            os.remove('output-latest.tgz')

        controller = Controller()
        controller.find_reporters()
        for d in controller.reports_dirs:
            if d and os.path.exists(d):
                print "purging contents of %s" % d
                shutil.rmtree(d)


        filters_dir = 'filters'
        # TODO improve this so it detects __init__.py customizations etc.
        if os.path.exists(filters_dir):
            if sorted(os.listdir(filters_dir)) == ['README', '__init__.py']:
                dexy_log.warn("purging contents of %s" % filters_dir)
                shutil.rmtree(filters_dir)
            else:
                print("Directory %s has been modified, not removing." % filters_dir)

    if args.reporters:
        args.run_dexy = False
        controller = Controller()
        reporters = controller.find_reporters()
        for r in reporters:
            print r.__name__
            if r.REPORTS_DIR:
                print "any generated reports will be saved in", r.REPORTS_DIR
            else:
                print "any generated reports will be saved in", args.logs_dir
            if r.__doc__:
                print r.__doc__
            else:
                print "no documentation available"
            print # finish with blank line
        print "Running reports can be disabled with the --no-reports option"

    if args.filters:
        args.run_dexy = False
        controller = Controller()
        filters = controller.find_filters()
        for k in sorted(filters.keys()):
            klass = filters[k]
            print
            print k, ":", klass.__name__
            if klass.executable():
                print "    calls", klass.executable()
            if klass.version_command():
                try:
                    raw_version = klass.version().strip()
                    if raw_version.find("\n") > 0:
                        # Clean up long version strings like R's
                        version = raw_version.split("\n")[0]
                    else:
                        version = raw_version

                    print "    version", version, "detected using",klass.version_command()
                except Exception: # TODO raise/catch specific Exception
                    print "    may not be installed, was unable to run", klass.version_command()
            if klass.__doc__:
                print "   ", klass.__doc__.strip()

    return args, dir_name, additional_excludes, dexy_log

def setup_controller():
    args, dir_name, additional_excludes, log = setup_option_parser()

    if not args.run_dexy:
        return None, args, log

    controller = Controller()
    controller.dir_name = dir_name
    controller.args = args
    controller.allow_remote = args.dangerous
    controller.artifact_class = args.artifact_class
    controller.artifacts_dir = args.artifacts_dir
    controller.logs_dir = args.logs_dir
    controller.log = log
    controller.config_file = args.config
    controller.use_local_files = args.local
    reporters = controller.find_reporters()
    if args.use_reporters:
        controller.reporters = [r for r in reporters if r.__name__ in args.use_reporters]
    else:
        controller.reporters = [r for r in reporters if r.DEFAULT]

    controller.additional_excludes = additional_excludes

    return controller, args, log

def dexy_command():
    controller, args, log = setup_controller()

    if not controller:
        return False

    if args.recurse:
        log.info("running dexy with recurse")
        for dirpath, dirnames, filenames in os.walk(controller.dir_name):
            process_dir = True

            if dirpath == ".":
                # We only exclude these dirs if they occur at project root level
                for x in EXCLUDE_DIRS_ROOT:
                    if x in dirnames:
                        log.debug("removing %s from list of child dirs %s of %s because it is in EXCLUDE_DIRS_ROOT" % (x, dirnames, dirpath))
                        dirnames.remove(x)

                for x in [controller.artifacts_dir, controller.logs_dir]:
                    if x in dirnames:
                        log.debug("removing %s from list of child dirs %s of %s because it is a dexy dir" % (x, dirnames, dirpath))
                        dirnames.remove(x)

                for x in controller.reports_dirs:
                    if x in dirnames:
                        log.debug("removing %s from list of child dirs %s of %s because it is a dexy report dir" % (x, dirnames, dirpath))
                        dirnames.remove(x)


            for x in EXCLUDE_DIRS_ALL_LEVELS:
                if x in dirnames:
                    log.debug("removing %s from list of child dirs %s of %s because it is in EXCLUDE_DIRS_ALL_LEVELS" % (x, dirnames, dirpath))
                    dirnames.remove(x)

            # If there is a file called .nodexy, this dir and all children are skipped.
            if os.path.isfile(os.path.join(dirpath, '.nodexy')):
                log.info("nodexy file found in %s" % dirpath)
                for d in dirnames:
                    # Remove all child dirs from processing.
                    dirnames.remove(d)
                process_dir = False

            # Now process any additional excludes specified on the command line
            for x in controller.additional_excludes:
                pattern = x
                m = re.match(pattern, dirpath)
                if not m:
                    pattern = "./%s" % x
                    m = re.match(pattern, dirpath)
                if m:
                    log.debug("removing %s because it matches pattern %s" % (dirpath, pattern))
                    for d in dirnames:
                        # Remove all child dirs from processing.
                        dirnames.remove(d)
                    process_dir = False
                else:
                    # check if children match pattern
                    for d in dirnames:
                        if x == d:
                            fullpath = os.path.join(dirpath, d)
                            log.debug("removing %s because it matches pattern %s" % (fullpath, pattern))
                            dirnames.remove(d)

            if not process_dir:
                log.warn("skipping dir %s" % dirpath)
            else:
                log.info("processing dir %s" % dirpath)
                controller.load_config(dirpath)
    else:
        log.info("not recursing")
        log.info("processing dir %s" % controller.dir_name)
        controller.load_config(controller.dir_name)

    if args.profile:
        cProfile.runctx("controller.setup_and_run()", globals(), locals(), "dexy.prof")
    else:
        controller.setup_and_run()

    if not args.no_reports:
        for reporter_klass in controller.reporters:
            reporter_klass(controller).run()
#            profile_memory("report-%s-complete" % reporter_klass.__name__)
    else:
        print 'reports not run'

    if args.serve:
        import SimpleHTTPServer
        import SocketServer

        os.chdir('output')

        PORT = 8000
        Handler = SimpleHTTPServer.SimpleHTTPRequestHandler

        httpd = SocketServer.TCPServer(("", PORT), Handler)

        print "serving at port", PORT
        httpd.serve_forever()




import kaa
import logging
from kaa import imlib2
from dexy.dexy_filter import DexyFilter
import os

class Imlib2Filter(DexyFilter):
    """Base class which reads previous artifact, calls a do_im method,
    saves the output. Can be inherited. Default action is to flip an image,
    pointless in itself but a useful demo to know things are working."""
    OUTPUT_EXTENSIONS = [".png"]
    INPUT_EXTENSIONS = [".png"]
    ALIASES = ['im-flipv']
    BINARY = True

    def do_im(self, image):
        image.flip_vertical()

    def process(self):
        # TODO fix font locations to be configurable
        imlib2.add_font_path("/usr/share/fonts/truetype/")
        imlib2.add_font_path("/usr/share/fonts/truetype/freefont/")
        self.default_font = imlib2.load_font("FreeSans", 20)

        wf = self.artifact.previous_artifact_filepath
        image = imlib2.open(wf)
        new_image = self.do_im(image)

        # Most of the time, we can just modify the image in do_im, but in some
        # cases we need to return a new image object, i.e. when cropping. So,
        # if do_im returns anything replace the image with the returned image.
        if new_image:
            image = new_image

        image.save(self.artifact.filepath())

class FlipHorizontalFilter(Imlib2Filter):
    ALIASES = ['im-fliph']

    def do_im(self, image):
        image.flip_horizontal()

class Imlib2GridFilter(Imlib2Filter):
    ALIASES = ['im-grid']

    def do_im(self, image):
        line_width = 3

        # draw vertical lines
        for i in xrange(100, image.width, 100):
            image.draw_rectangle((i, 0), (line_width, image.height), (0, 255, 0))
            image.draw_text((i+10, 0), "%s" % i, (255,0,0), self.default_font)

        # draw horizontal lines
        for j in xrange(100, image.height, 100):
            image.draw_rectangle((0, j), (image.width, line_width), (0, 0, 255))
            image.draw_text((0, j+10), "%s" % j, (255,0,0), self.default_font)

        for i in xrange(500, image.width, 500):
            for j in xrange(500, image.height, 500):
                image.draw_ellipse((i, j), (10, 10), (255,0,0))
                image.draw_text((i+15, j+15), "(%s,%s)" % (i, j), (255,0,0), self.default_font)

class Imlib2Crop(Imlib2Filter):
    ALIASES = ['im-crop']

    def do_im(self, image):
        # TODO specify crop region in params
        return image.crop((450, 450), (500, 500))

class Imlib2Thumb(Imlib2Filter):
    ALIASES = ['im-thumb']

    def do_im(self, image):
        if self.artifact.args.has_key('im-thumb'):
            region_size, start_x, start_y = self.artifact.args['im-thumb']['crop'].split("+")
            region_width, region_height = region_size.split("x")
            text = self.artifact.args['im-thumb']['title']
        else:
            self.log.debug("Using default cropping region and text for im-thumb, specify in args if you want something else")
            region_width=500
            region_height=500
            start_x=100
            start_y=100
            text="Example"

        # TODO specify crop region in params
        image = image.crop((int(start_x), int(start_y)), (int(region_width), int(region_height)))
        image = image.scale((500, 500))

        text_left = len(text)/2+30
        text_top = 10

        # Draw mock blank text to calculate height + width of text
        t_w, t_h, t_ha, t_va = image.draw_text((text_left, text_top), text, (255,255,255), self.default_font)

        # Draw white ellipse bigger than the ellipse people will see so it
        # doesn't run into neighboring text confusingly
        image.draw_ellipse((text_left+t_w/2, text_top+t_h/2), (t_w/2+25, t_h/2+20), (255,255,255), True)

        # Draw background ellipse
        image.draw_ellipse((text_left+t_w/2, text_top+t_h/2), (t_w/2+10, t_h/2+5), (238,221,130), True)

        # Draw actual text on top of ellipse
        image.draw_text((text_left, text_top), text, (0,0,0), self.default_font)

        return image




class Reporter(object):
    REPORTS_DIR = None
    DEFAULT = True # run this reporter by default

    def __init__(self, controller):
        self.controller = controller
        self.log = self.controller.log

    def run(self):
        pass



from dexy.reporter import Reporter
from dexy.utils import ansi_output_to_html
from jinja2 import Environment
from jinja2 import FileSystemLoader
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers.web import JavascriptLexer
import datetime
import json
import os
import shutil

class SiteReporter(Reporter):
    DEFAULT = False

    def __init__(self, controller):
        Reporter.__init__(self, controller)
        self.timestamp = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        self.report_dir = os.path.join(self.controller.logs_dir, "site-%s" % self.timestamp)
        self.latest_report_dir = os.path.join(self.controller.logs_dir, "site-latest")
        self.site_name = os.path.split(os.path.abspath(os.curdir))[1]
        self.pretty_site_name = " ".join([n.capitalize() for n in self.site_name.split("-")])

        # This shouldn't matter unless dexy gets run twice in 1 second
        shutil.rmtree(self.report_dir, ignore_errors=True)
        os.mkdir(self.report_dir)

    def setup_template(self):
        env = Environment()
        env.loader = FileSystemLoader(os.path.dirname(__file__))
        self.template = env.get_template('site_reporter_template.html')

    def setup_final_artifacts(self):
        self.artifacts = []

        for doc in self.controller.docs:
            artifact = doc.final_artifact()

            if artifact.final:
                self.artifacts.append(artifact)

            for k, a in artifact._inputs.items():
                if not a:
                    print "no artifact exists for key", k
                else:
                    if a.additional and a.final:
                        if not a.is_complete():
                            a.state = 'complete'
                            a.save()
                        self.artifacts.append(artifact)

    def setup_dirs(self):
        """Return a dict of directories and the artifacts in each directory."""
        self.dirs = {}
        for artifact in self.artifacts:
            dirname = os.path.dirname(artifact.canonical_filename())
            if self.dirs.has_key(dirname):
                if not artifact.key in [a.key for a in self.dirs[dirname]]:
                    self.dirs[dirname].append(artifact)
            else:
                self.dirs[dirname] = [artifact]

        self.html_filenames = {}
        self.artifact_filenames = {}
        for d in self.dirs.keys():
            self.html_filenames[d] = {}
            self.artifact_filenames[d] = {}

    def write_html_page_and_artifact(self, artifact):
        """Write artifact and maybe a HTML display page."""

        artifact_fn = artifact.canonical_filename()
        html_page_fn = "%s.html" % artifact_fn
        if artifact.ext == '.html':
            if not "|hd" in artifact.key: # TODO more robust test of whether full or fragment HTML
                artifact_fn = False
            elif os.path.basename(artifact_fn) == 'index.html':
                # we need index.html to be the html page
                html_page_fn = artifact_fn
                artifact_fn = artifact.long_canonical_filename()

        if artifact_fn:
            artifact.write_to_file(os.path.join(self.report_dir, artifact_fn))

        if artifact.ext in ['.html']:
            content = artifact.output_text()
        elif artifact.ext in ['.txt']:
            content = """<pre>\n%s\n</pre>""" % artifact.output_text()
        elif artifact.ext in ['.png', '.jpg', '.gif']:
            content = """<a href="%s"><img src="%s" /></a>""" % (artifact_fn, artifact_fn)
        elif artifact.ext in ['', '.pdf', '.css', '.rb', '.py', '.swf', '.R']:
            content = """<a href="%s">%s</a>""" % (artifact_fn, artifact_fn)
        else:
            self.log.debug("site reporter using default handling for extension %s" % artifact.ext)
            content = """<a href="%s">%s</a>""" % (artifact_fn, artifact_fn)

        env_data = {
            'a' : artifact,
            'title' : artifact.key,
            'content' : content,
            'directory' : os.path.dirname(html_page_fn),
            'dirs' : self.dirs.keys(),
            'pretty_site_name' : self.pretty_site_name,
            'relpath' : os.path.relpath,
            'basename' : os.path.basename,
            'site_name' : self.site_name,
            'artifact_fn' : artifact_fn,
            'html_page_fn' : html_page_fn
        }
        html_page_parent_dir = os.path.dirname(os.path.join(self.report_dir, html_page_fn))
        if not os.path.exists(html_page_parent_dir):
                os.makedirs(html_page_parent_dir)
        self.template.stream(env_data).dump(os.path.join(self.report_dir, html_page_fn))

    def run(self):
        self.setup_template()
        self.setup_final_artifacts()
        self.setup_dirs()

        for artifact in self.artifacts:
            self.write_html_page_and_artifact(artifact)

        def write_index(directory, children):
            # create an index page for each directory
            index_filename = os.path.join(self.report_dir, directory, 'index.html')
            env_data = {
                'children' : children,
                'directory' : directory,
                'content' : '',
                'dirs' : self.dirs.keys(),
                'pretty_site_name' : self.pretty_site_name,
                'relpath' : os.path.relpath,
                'basename' : os.path.basename,
                'site_name' : self.site_name
            }
            self.template.stream(env_data).dump(index_filename)


        for dirname, children in self.dirs.iteritems():
            write_index(dirname, children)

        shutil.rmtree(self.latest_report_dir, ignore_errors=True)
        shutil.copytree(self.report_dir, self.latest_report_dir)



from dexy.dexy_filter import DexyFilter
from dexy.utils import print_string_diff
from ordereddict import OrderedDict
import json
import os
import re
import shutil
import tarfile
import uuid
import zipfile

class ArchiveHandler(DexyFilter):
    """The archive handler creates .tgz archives of processed files. Create an
    empty/dummy file in the location you wish to have the archive."""
    OUTPUT_EXTENSIONS = [".tgz"]
    ALIASES = ['archive', 'tgz']
    BINARY = True
    FINAL = True

    def process(self):
        if self.artifact.args.has_key('use-short-names'):
            use_short_names = self.artifact.args['use-short-names']
        else:
            use_short_names = False
        af = self.artifact.filepath()
        tar = tarfile.open(af, mode="w:gz")
        for k, a in self.artifact.inputs().items():
            fn = a.filepath()
            if not os.path.exists(fn):
                raise Exception("File %s does not exist!" % fn)
            if use_short_names:
                arcname = a.canonical_filename()
            else:
                arcname = a.long_canonical_filename()
            self.artifact.log.debug("Adding file %s to archive %s." % (fn, af))
            tar.add(fn, arcname=arcname)
        tar.close()

class ZipArchiveHandler(DexyFilter):
    """The archive handler creates .zip archives of the input files. Create an
    empty file in the location you wish to have the archive."""
    OUTPUT_EXTENSIONS = [".zip"]
    ALIASES = ['zip']
    BINARY = True
    FINAL = True

    def process(self):
        if self.artifact.args.has_key('use-short-names'):
            use_short_names = self.artifact.args['use-short-names']
        else:
            use_short_names = False
        af = self.artifact.filepath()
        zf = zipfile.ZipFile(af, mode="w")
        for k, a in self.artifact.inputs().items():
            fn = a.filepath()
            if not os.path.exists(fn):
                raise Exception("File %s does not exist!" % fn)
            if use_short_names:
                arcname = a.canonical_filename()
            else:
                arcname = a.long_canonical_filename()
            self.artifact.log.debug("Adding file %s to archive %s." % (fn, af))
            zf.write(fn, arcname=arcname)
        zf.close()


class TestHandler(DexyFilter):
    """The test handler raises an error if output is not as expected. Handy for
    testing your custom filters or for ensuring that examples in your
    documentation stay correct."""

    ALIASES = ['test']

    def process(self):
        if not self.artifact.controller_args.run_tests:
            print "tests disabled, not running test", self.artifact.key
            self.artifact.data_dict = self.artifact.input_data_dict
            return False

        print "testing", self.artifact.key, "...",
        if not self.artifact.args.has_key('expects'):
            raise "You need to pass 'expects' to the test filter."

        expects = self.artifact.args['expects']
        if expects.startswith("@"):
            expects_file = open(expects.lstrip("@"), "r")
            expects_content = expects_file.read()
            expects_file.close()
        else:
            expects_content = expects
        expects = expects_content

        # TODO handle different types of expectation, e.g. when don't know exact
        # output but can look for type of data returned
        if hasattr(expects, 'items'):
            for k, v in self.artifact.input_data_dict.items():
                msg = print_string_diff(expects[k], v)
                msg += "\nexpected output '%s' (section %s) does not match actual '%s'" % (expects[k], k, v)
                assert expects[k] == v, msg
        else:
            inp = self.artifact.input_text()
            msg = print_string_diff(expects, inp)
            msg += "\nexpected output '%s' does not match actual '%s'" % (expects, inp)
            assert expects == inp, msg

        comment = ""
        if self.artifact.args.has_key("comment"):
           comment = "[%s]" % self.artifact.args['comment']
        print "ok", comment

        # Don't change the output so we can use end result still...
        self.artifact.data_dict = self.artifact.input_data_dict

class CopyHandler(DexyFilter):
    """
    Like 'dexy' filter for binary files. Copies the file without trying to read
    the contents. Hacky!
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['cp']
    BINARY = True
    FINAL = True

    def process(self):
        shutil.copyfile(self.artifact.name, self.artifact.filepath())

class JoinHandler(DexyFilter):
    """
    Takes sectioned code and joins it into a single section. Some filters which
    don't preserve sections will raise an error if they receive multiple
    sections as input, so this forces acknowledgement that sections will be
    lost.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['join']

    def process_dict(self, input_dict):
        return {'1' : self.artifact.input_text()}

class FooterHandler(DexyFilter):
    """
    Adds a footer to file. Looks for a file named _footer.ext where ext is the
    same extension as the file this is being applied to. So _footer.html for a
    file named index.html.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['ft', 'footer']

    def process_text(self, input_text):
        footer_key = "_footer%s" % self.artifact.ext
        footer_keys = []
        for k in self.artifact.inputs().keys():
            contains_footer = k.find(footer_key) > -1
            contains_pyg = k.find('|pyg') > 0
            if contains_footer and not contains_pyg:
                footer_keys.append(k)

        if self.artifact.args.has_key('footer'):
            requested_footer = self.artifact.args['footer']
            if not requested_footer in footer_keys:
                raise Exception("requested footer %s not found in %s" %
                                (requested_footer, ", ".join(footer_keys)))
            footer_keys = [requested_footer]

        if len(footer_keys) > 0:
            footer_key = sorted(footer_keys)[-1]
            self.artifact.log.debug("using %s as footer for %s" % (footer_key, self.artifact.key))
            footer_artifact = self.artifact.inputs()[footer_key]
            footer_text = footer_artifact.output_text()
        else:
            print self.artifact.inputs().keys()
            raise Exception("No file matching %s was found to work as a footer." % footer_key)

        return "%s\n%s" % (input_text, footer_text)

class HeaderHandler(DexyFilter):
    """
    Adds a header to file. Looks for a file named _header.ext where ext is the
    same extension as the file this is being applied to. So _header.html for a
    file named index.html.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['hd', 'header']

    def process_text(self, input_text):
        header_key = "_header%s" % self.artifact.ext
        header_keys = []
        for k in self.artifact.inputs().keys():
            contains_header = k.find(header_key) > -1
            contains_pyg = k.find('|pyg') > 0
            if contains_header and not contains_pyg:
                header_keys.append(k)

        if self.artifact.args.has_key('header'):
            requested_header = self.artifact.args['header']
            if not requested_header in header_keys:
                raise Exception("requested header %s not found in %s" %
                                (requested_header, ", ".join(header_keys)))
            header_keys = [requested_header]

        if len(header_keys) > 0:
            header_key = sorted(header_keys)[-1]
            self.artifact.log.debug("using %s as header for %s" % (header_key, self.artifact.key))
            header_artifact = self.artifact.inputs()[header_key]
            header_text = header_artifact.output_text()
        else:
            raise Exception("No file matching %s was found to work as a header for %s." % (header_key, self.artifact.key))

        return "%s\n%s" % (header_text, input_text)

# TODO implement combined header/footer handler as a shortcut

class HeadHandler(DexyFilter):
    """
    Returns just the first 10 lines of input.
    """
    ALIASES = ['head']
    def process_text(self, input_text):
        return "\n".join(input_text.split("\n")[0:10]) + "\n"

class WordWrapHandler(DexyFilter):
    """
    Wraps text after 79 characters (tries to preserve existing line breaks and
    spaces).
    """
    ALIASES = ['ww', 'wrap']

    #http://code.activestate.com/recipes/148061-one-liner-word-wrap-function/
    def wrap_text(self, text, width):
        """
        A word-wrap function that preserves existing line breaks
        and most spaces in the text. Expects that existing line
        breaks are posix newlines (\n).
        """
        return reduce(lambda line, word, width=width: '%s%s%s' %
                 (line,
                   ' \n'[(len(line)-line.rfind('\n')-1
                         + len(word.split('\n',1)[0]
                              ) >= width)],
                   word),
                  text.split(' ')
                 )

    def process_text(self, input_text):
        return self.wrap_text(input_text, 79)

class SplitHtmlHandler(DexyFilter):
    """Splits a HTML page into multiple HTML pages. The original page becomes an
    index page."""
    ALIASES = ['split', 'splithtml']
    INPUT_EXTENSIONS = [".html"]
    OUTPUT_EXTENSIONS = [".html"]
    FINAL = True

    def process(self):
        parent_dir = os.path.dirname(self.artifact.canonical_filename())
        input_text = self.artifact.input_text()

        if input_text.find("<!-- endsplit -->") > 0:
            body, footer = re.split("<!-- endsplit -->", input_text, maxsplit=1)
            sections = re.split("<!-- split \"(.+)\" -->\n", body)
            header = sections[0]

            pages = OrderedDict()
            index_content = None
            for i in range(1, len(sections), 2):
                if sections[i] == 'index':
                    index_content = sections[i+1]
                else:
                    section_name = sections[i]
                    # TODO proper url/filename escaping
                    section_url = section_name.replace(" ","-")

                    filename = "%s.html" % section_url
                    filepath = os.path.join(parent_dir, filename)
                    pages[section_name] = filename

                    artifact = self.artifact.add_additional_artifact(filepath, 'html')
                    artifact.set_data(header + sections[i+1] + footer)
                    artifact.save()

                    self.artifact.log.debug("added key %s to artifact %s ; links to file %s" %
                              (filepath, self.artifact.key, artifact.filename()))

            index_items = []
            for k in sorted(pages.keys()):
                index_items.append("""<li><a href="%s">%s</a></li>""" %
                                   (pages[k], k))

            output_dict = OrderedDict()
            output_dict['header'] = header
            if index_content:
                output_dict['index-page-content'] = index_content
            output_dict['index'] = "<ul>\n%s\n</ul>" % "\n".join(index_items)
            output_dict['footer'] = footer
        else:
            # No endsplit found, do nothing.
            output_dict = self.artifact.input_data_dict
        self.artifact.data_dict = output_dict

class SplitLatexHandler(DexyFilter):
    """Splits a latex doc into multiple latex docs."""
    ALIASES = ['splitlatex']
    INPUT_EXTENSIONS = [".tex"]
    OUTPUT_EXTENSIONS = [".tex"]
    FINAL = True

    def process(self):
        parent_dir = os.path.dirname(self.artifact.canonical_filename())
        input_text = self.artifact.input_text()

        if input_text.find("%% endsplit\n") > 0:
            body, footer = re.split("%% endsplit\n", input_text, maxsplit=1)
            sections = re.split("%% split \"(.+)\"\n", body)
            header = sections[0]

            pages = OrderedDict()
            for i in range(1, len(sections), 2):
                section_name = sections[i]
                # TODO proper url/filename escaping
                section_url = section_name.replace(" ","-")
                source = header + sections[i+1] + footer

                ext = '.tex'

                filename = "%s%s" % (section_url, ext)
                filepath = os.path.join(parent_dir, filename)
                pages[section_name] = filename

                artifact = self.artifact.__class__(filepath)
                artifact.ext = ext
                artifact.binary = False
                artifact.final = True
                artifact.artifacts_dir = self.artifact.artifacts_dir
                artifact.hashstring = str(uuid.uuid4())
                artifact.set_data(source)
                artifact.save()

                self.artifact.inputs()[filepath] = artifact
                self.artifact.log.debug("added key %s to artifact %s ; links to file %s" %
                          (filepath, self.artifact.key, artifact.filename()))

            index_items = []
            for k in sorted(pages.keys()):
                index_items.append("""<li><a href="%s">%s</a></li>""" %
                                   (pages[k], k))

        output_dict = self.artifact.input_data_dict
        self.artifact.data_dict = output_dict

class PrettyPrintJsonFilter(DexyFilter):
    ALIASES = ['ppjson']
    OUTPUT_EXTENSIONS = ['.json']

    def process_text(self, input_text):
        json_content = json.loads(input_text)
        return json.dumps(json_content, sort_keys=True, indent=4)

class SillyHandler(DexyFilter):
    ALIASES =['silly']

    def process_text(self, input_text):
        return "you said: '%s'\n that's silly!\n" % input_text



from dexy.dexy_filter import DexyFilter
import handlers
import shutil

class RFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .R extension if able.
    """
    INPUT_EXTENSIONS = [".R"]
    OUTPUT_EXTENSIONS = [".R"]
    ALIASES = ['forcer']

class PdfFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .pdf extension if able.
    """
    INPUT_EXTENSIONS = [".pdf"]
    OUTPUT_EXTENSIONS = [".pdf"]
    ALIASES = ['p', 'forcepdf']
    BINARY = True

    def process(self):
        shutil.copyfile(self.artifact.previous_artifact_filepath, self.artifact.filepath())

class ForceJpgFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .png extension if able.
    """
    INPUT_EXTENSIONS = [".jpg"]
    OUTPUT_EXTENSIONS = [".jpg"]
    ALIASES = ['jn', 'forcejpg']
    BINARY = True

    def process(self):
        shutil.copyfile(self.artifact.previous_artifact_filepath, self.artifact.filepath())

class ForcePngFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .png extension if able.
    """
    INPUT_EXTENSIONS = [".png"]
    OUTPUT_EXTENSIONS = [".png"]
    ALIASES = ['pn', 'forcepng']
    BINARY = True

    def process(self):
        shutil.copyfile(self.artifact.previous_artifact_filepath, self.artifact.filepath())

class ConvertBashFormatHandler(DexyFilter):
    """
    Converts whatever file extension is input to be .bash.
    """
    INPUT_EXTENSIONS = [".*", "*"]
    OUTPUT_EXTENSIONS = [".sh"]
    ALIASES = ['b', 'forcebash']

class ConvertTextFormatHandler(DexyFilter):
    """
    Changes whatever file extension is input to be .txt.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['ct']

class ConvertHTMLFormatHandler(DexyFilter):
    """
    Changes whatever file extension is input to be .html.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['ch']

class TextFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .txt extension if able.
    """
    INPUT_EXTENSIONS = [".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['t', 'forcetext']

class XmlFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .xml extension if able.
    """
    INPUT_EXTENSIONS = [".xml"]
    OUTPUT_EXTENSIONS = [".xml"]
    ALIASES = ['x', 'forcexml']

class LatexFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .tex extension if able.
    """
    INPUT_EXTENSIONS = [".tex"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['l', 'forcelatex']

class HtmlFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .html extension if able.
    """
    INPUT_EXTENSIONS = [".html"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['h', 'forcehtml']

class JsonFormatHandler(DexyFilter):
    """
    Does nothing, just forces previous filter to output .json extension if able.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".json"]
    ALIASES = ['j', 'forcejson']




from dexy.dexy_filter import DexyFilter
from jinja2 import Environment
from jinja2 import StrictUndefined
from ordereddict import OrderedDict
import jinja2
import json
import os
import re
import traceback
import urllib
import pprint

try:
    import pygments
    import pygments.formatters
    from pygments.styles import get_all_styles
    USE_PYGMENTS = True
except ImportError:
    USE_PYGMENTS = False

class FilenameHandler(DexyFilter):
    """Generate random filenames to track provenance of data."""
    ALIASES = ['fn']

    def process_text(self, input_text):
        # TODO this should not match more than two dashes
        # TODO differentiate between using the same base name for 2 extensions, or at least warn
        for m in re.finditer("dexy--(\S+)\.([a-z]+)", input_text):
            key = m.groups()[0]
            ext = m.groups()[1]
            key_with_ext = "%s.%s" % (key, ext)
            key_with_ext_with_dexy = "%s|dexy" % key_with_ext

            if key_with_ext in self.artifact.inputs().keys():
                artifact = self.artifact.inputs()[key_with_ext]
                self.artifact.log.debug("[fn] existing key %s in artifact %s links to file %s" %
                          (key_with_ext, self.artifact.key, artifact.filename()))
            elif key_with_ext_with_dexy in self.artifact.inputs().keys():
                artifact = self.artifact.inputs()[key_with_ext_with_dexy]
                self.artifact.log.debug("[fn] existing key %s in artifact %s links to existing file %s" %
                          (key_with_ext, self.artifact.key, artifact.filename()))
            else:
                artifact = self.artifact.add_additional_artifact(key_with_ext, ext)
                self.artifact.log.debug("[fn] added key %s to artifact %s ; links to new file %s" %
                          (key_with_ext, self.artifact.key, artifact.filename()))

            input_text = input_text.replace(m.group(), artifact.filename())

        # Hack to replace __ with -- in case we want to document how to use this
        # filter, we can't use -- because this will be acted upon.
        for m in re.finditer("dexy__(.+)\.([a-z]+)", input_text):
            key = m.groups()[0]
            ext = m.groups()[1]
            replacement_key = "dexy--%s.%s" % (key, ext)
            input_text = input_text.replace(m.group(), replacement_key)

        return input_text

class JinjaHandler(DexyFilter):
    """
    Runs the Jinja templating engine on your document to incorporate dynamic
    content.
    """
    ALIASES = ['jinja']
    FINAL = True
    CHANGE_EXTENSIONS = {
        ".mds" : ".md"
    }

    def pre_and_clippy(self, text):
        return """<pre>
%s
</pre>%s
""" % (text, self.clippy_helper(text))

    def clippy_helper(self, text):
        if not text or len(text) == 0:
            raise Exception("You passed blank text to clippy helper!")
        quoted_text = urllib.quote(text)
        return """<object classid="clsid:d27cdb6e-ae6d-11cf-96b8-444553540000"
                width="110"
                height="14"
                id="clippy" >
        <param name="movie" value="/clippy.swf"/>
        <param name="allowScriptAccess" value="always" />
        <param name="quality" value="high" />
        <param name="scale" value="noscale" />
        <param NAME="FlashVars" value="text=%s">
        <param name="bgcolor" value="#ffffff">
        <embed src="/clippy.swf"
               width="110"
               height="14"
               name="clippy"
               quality="high"
               allowScriptAccess="always"
               type="application/x-shockwave-flash"
               pluginspage="http://www.macromedia.com/go/getflashplayer"
               FlashVars="text=%s"
               bgcolor="#ffffff"
        />
        </object>""" % (quoted_text, quoted_text)

    def process_text(self, input_text):
        if self.artifact.ext in self.CHANGE_EXTENSIONS.keys():
            self.artifact.ext = self.CHANGE_EXTENSIONS[self.artifact.ext]

        document_data = {}
        raw_data = {}
        local_inputs = {} # docs in the same directory
        child_inputs = {} # inputs beneath this doc

        artifact_dir = os.path.join(os.curdir, os.path.dirname(self.artifact.name))
        subdirectories = [d for d in os.listdir(artifact_dir) if os.path.isdir(os.path.join(artifact_dir, d))]

        notextile = self.artifact.args.has_key('notextile') and self.artifact.args['notextile']

        for key, a in self.artifact.inputs().items():
            if a.is_output_cached():
                self.artifact.log.debug("Loading artifact %s" % key)
                a.load() # reload

            # Full path keys
            keys = [key, a.canonical_filename()]


            artifact_dir = os.path.dirname(self.artifact.name)
            key_dir = os.path.dirname(key)
            # Shortcut keys if in common directory
            if artifact_dir == key_dir or not key_dir:
                keys.append(os.path.basename(key))
                fn = a.canonical_filename()
                keys.append(os.path.basename(fn))
                keys.append(os.path.splitext(fn)[0]) # TODO deal with collisions
                if key_dir or not artifact_dir:
                    # This is also a local input.
                    local_inputs[key] = a

            elif key_dir and artifact_dir and os.path.relpath(artifact_dir, key_dir) in ["..", "../.."]:
                child_inputs[key] = a

            # Do special handling of data
            if a.ext == '.json':
                path_to_file = os.path.join(self.artifact.artifacts_dir, a.filename())
#                if not os.path.exists(path_to_file):
#                    a.state = 'complete'
#                    a.save()
                # TODO read this from memory rather than loading from file?
                # TODO capture load errors
                try:
                    unsorted_json = json.load(open(path_to_file, "r"))
                except ValueError as e:
                    print a.key
                    raise e

                def sort_dict(d):
                    od = OrderedDict()
                    for k in sorted(d.keys()):
                        v = d[k]
                        if isinstance(v, dict) or isinstance(v, OrderedDict):
                            od[k] = sort_dict(v)
                        else:
                            if notextile and isinstance(v, str) and "<span" in v and not "<notextile>" in v:
                                od[k] = "<notextile>\n%s\n</notextile>" % v
                            else:
                                od[k] = v
                    return od

                if hasattr(unsorted_json, 'keys'):
                    data = sort_dict(unsorted_json)
                else:
                    data = unsorted_json

            elif notextile and a.ext == '.html':
                data = OrderedDict()
                for s, h in a.data_dict.items():
                    if "<notextile>" in h:
                        data[s] = h
                    else:
                        data[s] = "<notextile>\n%s\n</notextile>" % h
            else:
                data = a.data_dict

            for k in keys:
                raw_data[k] = a.output_text()
                if data.keys() == ['1']:
                    document_data[k] = data['1']
                else:
                    document_data[k] = data

        if self.artifact.ext == ".tex":
            self.artifact.log.debug("changing jinja tags to << >> etc. for %s" % self.artifact.key)
            env = Environment(
                block_start_string = '<%',
                block_end_string = '%>',
                variable_start_string = '<<',
                variable_end_string = '>>',
                comment_start_string = '<#',
                comment_end_string = '#>',
                undefined = StrictUndefined
                )
        else:
            env = Environment()

        template_hash = {
            's' : self.artifact,
            'f' : self,
            'a' : self.artifact._inputs,
            'c' : child_inputs,
            'ck' : sorted(child_inputs.keys()),
            'ak' : sorted(self.artifact._inputs.keys()),
            'd' : document_data,
            'dk' : sorted(document_data.keys()),
            'l' : local_inputs,
            'lk' : sorted(local_inputs.keys()),
            'r' : raw_data,
            'subdirectories' : subdirectories,
            'len' : len,
            'pformat' : pprint.pformat,
            'WARN_AUTOGEN' : "This document is generated using Dexy (http://dexy.it). You should modify the source files, not this generated output."
        }


        # Create pygments stylesheets
        if USE_PYGMENTS:
            if not hasattr(self.__class__, 'PYGMENTS_STYLESHEETS'):
                pygments_stylesheets = {}
                for style_name in get_all_styles():
                    for formatter_class in [pygments.formatters.LatexFormatter, pygments.formatters.HtmlFormatter]:
                        pygments_formatter = formatter_class(style=style_name)
                        style_info = pygments_formatter.get_style_defs()

                        for fn in pygments_formatter.filenames:
                            ext = fn.split(".")[1]
                            key = "%s.%s" % (style_name, ext)
                            pygments_stylesheets[key] = style_info

                # cache this so we only need to do this once per dexy run...
                self.__class__.PYGMENTS_STYLESHEETS = pygments_stylesheets
            template_hash['pygments'] = self.__class__.PYGMENTS_STYLESHEETS

        if self.artifact.args.has_key('jinjavars'):
            for k, v in self.artifact.args['jinjavars'].items():
                if template_hash.has_key(k):
                    raise Exception("Please do not set a jinjavar for %s as this conflicts with standard vars: %s" % (k, ', '.join(template_hash.keys())))
                template_hash[k] = v

        if self.artifact.controller_args.globals:
            for k, v in self.artifact.controller_args.globals.iteritems():
                if not template_hash.has_key(k):
                    template_hash[k] = v

        try:
            template = env.from_string(input_text)
            result = str(template.render(template_hash))
        except jinja2.exceptions.TemplateSyntaxError as e:
            print "*" * 80
            print "there is a problem with line", e.lineno, "of your file %s" % self.artifact.name
            print e.message
            input_lines = self.artifact.input_text().splitlines()
            print "=" * 60
            if e.lineno >= 3:
                print "   ", input_lines[e.lineno-3]
            if e.lineno >= 2:
                print "   ", input_lines[e.lineno-2]
            print ">>>", input_lines[e.lineno-1]
            if len(input_lines) >= e.lineno:
                print "   ", input_lines[e.lineno-0]
            if len(input_lines) >= (e.lineno + 1):
                print "   ", input_lines[e.lineno+1]
            print "=" * 60
            print "^" * 80
            result = ""
        except Exception as e:
            print "*" * 80
            print "there is a problem with your file %s" % self.artifact.name
            print e.message
            if len(self.artifact.input_text().splitlines()) < 20:
                print "=" * 60
                print self.artifact.input_text()
                print "=" * 60
            else:
                print "Please check your file, you may have mistyped a variable name."
            print
            print "Here is a traceback which may have more information about the problem with your file %s" % self.artifact.name
            traceback.print_exc()
            print "^" * 80
            result = ""

        return result
